static char rcsid[] = "$Id: comm.c,v 1.68 1995/08/21 20:55:10 hardy Exp $";
/* 
 * File:         comm.c
 * Description:  socket-based communication facility.  Adapted from DHT
 *               authenticating message library.
 * Author:       John Noll, USC
 * Created:      Tue Apr 26 14:24:37 1994
 * Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */

#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>		/* For MAXHOSTNAMELEN */
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#ifdef _AIX
#include <sys/select.h>
#endif

#include "debug.h"
#include "comm.h"
#include "ipcache.h"
#include "cache_cf.h"
#include "fdstat.h"

#define min(x,y) ((x)<(y)? (x) : (y))

FD_ENTRY *fd_table;
int *fd_lifetime;
time_t cached_curtime = 0L;	/* global time var set by select loop */

static fd_set send_sockets;
static fd_set receive_sockets;

static int (*app_handler) ();

extern int errno;
extern int do_reuse;
#if !defined(__FreeBSD__)
extern char *sys_errlist[];
#endif
extern int getMaxFD();

static void comm_handler()
{
    /* Call application installed handler. */
    debug(5, "comm_handler:\n");
    app_handler();
}

char *comm_hostname()
{
    static char host[MAXHOSTNAMELEN + 1];
    static int present = 0;
    struct hostent *h = NULL;

    /* Get the host name and store it in host to return */
    if (!present) {
	host[0] = '\0';
	if (gethostname(host, MAXHOSTNAMELEN) == -1) {
	    debug(1, "comm_hostname: gethostname failed: %s\n",
		  sys_errlist[errno]);
	    return NULL;
	} else {
	    if ((h = ipcache_gethostbyname(host)) != NULL) {
		/* DNS lookup successful */
		/* use the official name from DNS lookup */
		strcpy(host, h->h_name);
	    }
	    present = 1;
	}
    }
    return host;
}

u_num32 comm_hostip()
{
	static long ip=0;
	struct hostent *h = 0;

	if (ip)
	{
		return ip;
	}
	else
	{
		/* Get the host name */
		if ((h = ipcache_gethostbyname(comm_hostname())) != NULL)
		{
			/* Use the first ip address of this machine */
			memcpy(&ip,h->h_addr_list[0], 
				(h->h_length>sizeof(u_num32))?sizeof(u_num32):h->h_length) ;
			return ip;
		}
		else
		{
			/* XXX Log a problem */
			return 0;
		}
	}
}

char *comm_hostname_direct()
{
    static char temp_host[MAXHOSTNAMELEN + 1];

    temp_host[0] = '\0';
    if (gethostname(temp_host, MAXHOSTNAMELEN) == -1) {
	debug(1, "comm_hostname_direct: gethostname failed: %s\n",
	      sys_errlist[errno]);
	return NULL;
    } else {
	return temp_host;
    }
}

/* Return the local port associated with fd. */
int comm_port(fd)
int fd;
{
    if (fd_table[fd].port) {
	return fd_table[fd].port;
    } else {
	struct sockaddr_in addr;
	int addr_len = sizeof(addr);

	if (getsockname(fd, (struct sockaddr *) &addr, &addr_len)) {
	    debug(1, "comm_port: Failed to retrieve TCP/UDP port number for socket: FD %d: %s\n", fd, sys_errlist[errno]);
	    return (COMM_ERROR);
	}
	debug(6, "comm_port: FD %d: sockaddr %u.\n", fd, addr.sin_addr.s_addr);
	fd_table[fd].port = ntohs(addr.sin_port);

	return fd_table[fd].port;
    }
}

/* Create a socket. Default is blocking, stream (TCP) socket.  IO_TYPE
   is OR of flags specified in comm.h. */
int comm_open(io_type, port, handler)
unsigned int io_type;
int port;
int (*handler) ();		/* Interrupt handler. */
{
    struct sockaddr_in addr;
    struct in_addr *getAddress();
    int io_handler();
    int new_socket;
    FD_ENTRY *conn;
    int sock_type = io_type & COMM_DGRAM ? SOCK_DGRAM : SOCK_STREAM;

    /* Create socket for accepting new connections. */
    if ((new_socket = socket(AF_INET, sock_type, 0)) < 0) {
	debug(0, "comm_open: socket failure: %s\n", sys_errlist[errno]);
	return (COMM_ERROR);
    }
    /* update fdstat */
    fdstat_open(new_socket, Socket);

    conn = &fd_table[new_socket];
    memset(conn, '\0', sizeof(FD_ENTRY));
    conn->openned = 1;

    /* Create name for the socket */
    memset(&addr, '\0', sizeof(addr));
    addr.sin_family = AF_INET;

    /*
     * If port is nonzero, it is the port number specified by user.
     */

    if (getAddress(0)) {
	memcpy(&addr.sin_addr, getAddress(0), sizeof(addr.sin_addr));
    } else {
	addr.sin_addr.s_addr = INADDR_ANY;
    }

    addr.sin_port = htons(port);

    if (port > 0) {
	setsocketNolinger(new_socket);
	if (do_reuse) {
	    setsocketReuse(new_socket);
	}
    }
    if (port && bind(new_socket, (struct sockaddr *) &addr, sizeof(addr))) {
	debug(0, "comm_open: bind failure on addr %s port %d: %s\n",
	      inet_ntoa(addr.sin_addr), port, sys_errlist[errno]);
	return (COMM_ERROR);
    }
    conn->port = port;

    if (io_type & COMM_INTERRUPT) {
#ifdef __hpux
	pid_t my_pid = getpid();
	int non_zero = 1;

	/* Install interrupt handler. */
	if ((app_handler = conn->handler = handler) == NULL)
	    return (COMM_NO_HANDLER);

	/* Set up the environment for handling signals. */
	signal(SIGIO, comm_handler);

	/* Set the process receiving SIGIO/SIGURG signals to us. */
	if (ioctl(new_socket, SIOCSPGRP, &my_pid) < 0) {
	    debug(0, "comm_open: Failure to set SIOCSPGRP: %s\n",
		  sys_errlist[errno]);
	    return (COMM_ERROR);
	}
	/*
	 * Set status flags to allow receipt of asychronous I/O signals.
	 */

	if (ioctl(new_socket, FIOASYNC, &non_zero)) {
	    debug(0, "comm_open: Failure to set FIOASYNC: %s\n",
		  sys_errlist[errno]);
	    return COMM_ERROR;
	}
#else
	/* Install interrupt handler. */
	if ((app_handler = conn->handler = handler) == NULL)
	    return (COMM_NO_HANDLER);

	/* Set up the environment for handling signals. */
	signal(SIGIO, comm_handler);

	/* Set the process receiving SIGIO/SIGURG signals to us. */
	if (fcntl(new_socket, F_SETOWN, getpid()) < 0) {
	    debug(0, "comm_open: Failure to set F_SETOWN: %s\n",
		  sys_errlist[errno]);
	    return (COMM_ERROR);
	}
	/* Set status flags to allow receipt of asychronous I/O signals.  */
	if (fcntl(new_socket, F_SETFL, FASYNC)) {
	    debug(0, "comm_open: Failure to set FNDELAY | FASYNC: %s\n",
		  sys_errlist[errno]);
	    return COMM_ERROR;
	}
#endif				/* HPUX */
    }
    if (io_type & COMM_NONBLOCKING) {
#ifdef __hpux
	/*
	 * Set up the flag NOT to have the socket to wait for message from
	 * network forever, but to return -1 when no message is coming in.
	 */
	if (fcntl(new_socket, F_SETFL, O_NONBLOCK)) {
	    debug(0, "comm_open: Failure to set O_NONBLOCK: %s\n",
		  sys_errlist[errno]);
	    return (COMM_ERROR);
	}
#else
	/*
	 * Set up the flag NOT to have the socket to wait for message from
	 * network forever, but to return -1 when no message is coming in.
	 */
	if (fcntl(new_socket, F_SETFL, FNDELAY)) {
	    debug(0, "comm_open: Failure to set FNDELAY: %s\n",
		  sys_errlist[errno]);
	    return (COMM_ERROR);
	}
#endif				/* HPUX */
    }
    conn->comm_type = io_type;
    return new_socket;
}

/*      NOTE: set the listen queue to 50 and rely on the kernel to      **
   **   impose an upper limit.  Solaris' listen(3n) page says it has    **
   **   no limit on this parameter, but sys/socket.h sets SOMAXCONN     **
   **   to 5.  HP-UX currently has a limit of 20.  SunOS is 5 and       **
   **   OSF 3.0 is 8.                                                   */

/* Listen on SOCK. */
int comm_listen(sock)
int sock;
{
    int x;
    FD_SET(sock, &receive_sockets);
    if ((x = listen(sock, 50)) < 0) {
	debug(0, "comm_listen: listen(%d, 50): %s\n",
	      sock, sys_errlist[errno]);
	return x;
    }
    return sock;
}


/* Connect SOCK to specified DEST_PORT at DEST_HOST. */
int comm_connect(sock, dest_host, dest_port)
int sock;			/* Type of communication to use. */
char *dest_host;		/* Server's host name. */
int dest_port;			/* Server's port. */
{
    struct hostent *hp;
    static struct sockaddr_in to_addr;

    /* Set up the destination socket address for message to send to. */
    to_addr.sin_family = AF_INET;

    if ((hp = ipcache_gethostbyname(dest_host)) == 0) {
	debug(1, "comm_connect: Failure to lookup host: %s.\n", dest_host);
	return (COMM_ERROR);
    }
    memcpy(&to_addr.sin_addr, hp->h_addr, hp->h_length);
    to_addr.sin_port = htons(dest_port);
    return comm_connect_addr(sock, &to_addr);
}

int comm_connect_addr(sock, address)
int sock;
struct sockaddr_in *address;
{
    int status = COMM_OK;
    FD_ENTRY *conn = &fd_table[sock];
    int len, x;

    /* sanity check */
    if (ntohs(address->sin_port) <= 0) {
	debug(10, "comm_connect_addr: %s:%d: URL uses port 0?\n",
	      inet_ntoa(address->sin_addr), ntohs(address->sin_port));
	errno = 0;
	return COMM_ERROR;
    }
    /* Establish connection. */
    if (connect(sock, (struct sockaddr *) address, sizeof(struct sockaddr_in)) < 0)
	switch (errno) {
	case EALREADY:
	    return COMM_ERROR;
	    break;
	case EINPROGRESS:
	    status = EINPROGRESS;
	    break;
	case EISCONN:
	    status = COMM_OK;
	    break;
	case EINVAL:
	    len = sizeof(x);
	    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (char *) &x, &len) >= 0)
		errno = x;
	default:
	    debug(1, "comm_connect_addr: %s:%d: socket failure: %s.\n",
		  inet_ntoa(address->sin_addr),
		  ntohs(address->sin_port),
		  sys_errlist[errno]);
	    return COMM_ERROR;
	}
    /* set the lifetime for this client */
    fd_lifetime[sock] = time(NULL) + getClientLifetime();
    debug(10, "comm_connect_addr: FD %d (lifetime %d): connected to %s:%d.\n",
	  sock, fd_lifetime[sock], inet_ntoa(address->sin_addr),
	  ntohs(address->sin_port));

    /* Add new socket to list of open sockets. */
    FD_SET(sock, &send_sockets);
    conn->sender = 1;

    return status;
}

/* Wait for an incoming connection on FD.  FD should be a socket returned
   from comm_listen. */
int comm_accept(fd)
int fd;
{
    int sock;
    struct sockaddr_in from;	/* address of sender */
    int fromlen;
    FD_ENTRY *conn;
    FD_ENTRY *listener = &fd_table[fd];

    fromlen = sizeof(from);
    while ((sock = accept(fd, (struct sockaddr *) &from, &fromlen)) < 0) {
	if (errno == EWOULDBLOCK)
	    return COMM_NOMESSAGE;
	else if (errno != EINTR) {
	    debug(1, "comm_accept: accept failure: %s\n", sys_errlist[errno]);
	    return COMM_ERROR;
	}
	/* if accept interrupted, try again */
    }

    /* fdstat update */
    fdstat_open(sock, Socket);

    conn = &fd_table[sock];
    conn->openned = 1;
    conn->sender = 0;		/* This is an accept, therefore receiver. */
    conn->comm_type = listener->comm_type;

    FD_SET(sock, &receive_sockets);

    return sock;
}

int comm_close(fd)
int fd;
{
    FD_ENTRY *conn = &fd_table[fd];

    FD_CLR(fd, &receive_sockets);
    FD_CLR(fd, &send_sockets);

    fd_lifetime[fd] = -1;	/* invalidate the lifetime */
    debug(10, "comm_close: FD %d\n", fd);
    /* update fdstat */
    fdstat_close(fd);
    memset(conn, '\0', sizeof(FD_ENTRY));
    return close(fd);
}

/* use to clean up fdtable when socket is closed without
   using comm_close */
int comm_cleanup_fd_entry(fd)
int fd;
{
    FD_ENTRY *conn = &fd_table[fd];

    memset(conn, 0, sizeof(FD_ENTRY));
    return 0;
}


/* Send a udp datagram to specified PORT at HOST. */
int comm_udp_send(fd, host, port, buf, len)
int fd;
char *host;
int port;
char *buf;
int len;
{
    struct hostent *hp;
    static struct sockaddr_in to_addr;
    int bytes_sent;

    /* Set up the destination socket address for message to send to. */
    to_addr.sin_family = AF_INET;

    if ((hp = ipcache_gethostbyname(host)) == 0) {
	debug(1, "comm_udp_send: gethostbyname failure: %s: %s\n",
	      host, sys_errlist[errno]);
	return (COMM_ERROR);
    }
    memcpy(&to_addr.sin_addr, hp->h_addr, hp->h_length);
    to_addr.sin_port = htons(port);
    if ((bytes_sent = sendto(fd, buf, len, 0, (struct sockaddr *) &to_addr,
			     sizeof(to_addr))) < 0) {
	debug(1, "comm_udp_send: sendto failure: FD %d: %s\n",
	      fd, sys_errlist[errno]);
	return COMM_ERROR;
    }
    return bytes_sent;
}

/* Send a udp datagram to specified TO_ADDR. */
int comm_udp_sendto(fd, to_addr, addr_len, buf, len)
int fd;
struct sockaddr_in *to_addr;
int addr_len;
char *buf;
int len;
{
    int bytes_sent;

    if ((bytes_sent = sendto(fd, buf, len, 0, (struct sockaddr *) to_addr, addr_len)) < 0) {
	debug(1, "comm_udp_sendto: sendto failure: FD %d: %s\n",
	      fd, sys_errlist[errno]);
	return COMM_ERROR;
    }
    return bytes_sent;
}

int comm_udp_recv(fd, buf, size, from_addr, from_size)
int fd;
char *buf;
int size;
struct sockaddr_in *from_addr;
int *from_size;			/* in: size of from_addr; out: size filled in. */
{
    int len = recvfrom(fd, buf, size, 0, (struct sockaddr *) from_addr,
		       from_size);
    if (len < 0) {
	debug(1, "comm_udp_recv: recvfrom failure: FD %d: %s\n", fd,
	      sys_errlist[errno]);
	return COMM_ERROR;
    }
    return len;
}



/* Select on all sockets; call handlers for those that are ready. */
int comm_select(sec, usec, failtime)
long sec, usec;
time_t failtime;
{
    int fd, i, max_fd = getMaxFD();
    fd_set readfds, writefds, exceptfds;
    int num;
    struct timeval timeout, poll_time;


    /* assume all process are very fast under 1 second. Do  time() only once */
    cached_curtime = time(0L);
    /* use only 1 second granularity */
    timeout.tv_sec = cached_curtime + sec;
    timeout.tv_usec = usec;


    while ((timeout.tv_sec > (cached_curtime = time(0L))) &&
	   (cached_curtime < failtime)) {

	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	FD_ZERO(&exceptfds);

	for (i = 0; i < fdstat_biggest_fd() + 1; i++) {
	    /* Check each open socket for a handler. */
	    if (fd_table[i].read_handler)
		FD_SET(i, &readfds);
	    if (fd_table[i].write_handler)
		FD_SET(i, &writefds);
	    if (fd_table[i].except_handler)
		FD_SET(i, &exceptfds);
	}

	poll_time.tv_sec = 1;
	poll_time.tv_usec = 0;

	while ((num = select(fdstat_biggest_fd() + 1,
		     &readfds, &writefds, &exceptfds, &poll_time)) < 0) {

	    if (errno != EINTR) {
		debug(0, "comm_select: select failure: %s (errno %d).\n",
		      sys_errlist[errno], errno);
		examine_select(&readfds, &writefds, &exceptfds);
		return COMM_ERROR;
	    }
	    /* if select interrupted, try again */
	}

	debug(num ? 5 : 8, "comm_select - %d sockets ready at %d\n",
	      num, cached_curtime);

	/* scan for timeout */
	for (fd = 0; fd < (fdstat_biggest_fd() + 1); ++fd) {
	    if ((fd_table[fd].timeout_handler) &&
		(fd_table[fd].timeout_time <= cached_curtime)) {
		int (*tmp) () = fd_table[fd].timeout_handler;
		debug(5, "comm_select - timeout on socket %d at %d\n",
		      fd, cached_curtime);
		fd_table[fd].timeout_handler = 0;
		tmp(fd, fd_table[fd].timeout_data);
	    }
	}

	/* scan for hardwired lifetime expires, do the timeouts first though */
	for (fd = 0; fd < max_fd; fd++) {
	    if ((fd_lifetime[fd] != -1) && (fd_lifetime[fd] < cached_curtime)) {
		int use_lifetime_handler = 0;
		int use_read = 0;
		int (*tmp_local) () = NULL;

		if (fd_table[fd].lifetime_handler != NULL) {
		    use_lifetime_handler = 1;
		    tmp_local = fd_table[fd].lifetime_handler;
		    fd_table[fd].lifetime_handler = 0;	/* reset it */
		} else if (fd_table[fd].read_handler != NULL) {
		    use_read = 1;
		    tmp_local = fd_table[fd].read_handler;
		    fd_table[fd].read_handler = 0;	/* reset it */
		} else if (fd_table[fd].write_handler != NULL) {
		    use_read = 0;
		    tmp_local = fd_table[fd].write_handler;
		    fd_table[fd].write_handler = 0;	/* reset it */
		} else {
		    use_read = 0;
		    tmp_local = NULL;
		}
		if (tmp_local) {
		    if (use_lifetime_handler) {
			debug(2, "comm_select - %d lifetime expire: %d < %d (Lifetime handler %p)\n",
			 fd, fd_lifetime[fd], cached_curtime, tmp_local);
		    } else {
			debug(2, "comm_select - %d lifetime expire: %d < %d (%s handler %p)\n",
			      fd, fd_lifetime[fd], cached_curtime,
			      use_read ? "read" : "write", tmp_local);
		    }
		} else {
		    debug(1, "comm_select - %d lifetime expire: %d < %d (handler not available.)\n",
			  fd, fd_lifetime[fd], cached_curtime);
		}

		if (tmp_local != NULL) {
		    if (use_lifetime_handler) {
			tmp_local(fd, fd_table[fd].lifetime_data);
		    } else {
			/* 
			 *  we close(2) first so that the handler fails and 
			 *  deallocates the structure.
			 */
			(void) close(fd);
			tmp_local(fd, use_read ? fd_table[fd].read_data :
				  fd_table[fd].write_data);
		    }
		    if (fd_table[fd].openned) {
			/* hmm.. still openned. do full comm_close */
			debug(5, "comm_select - %d lifetime expire: %d < %d : Handler did not close the socket.\n comm_select will do.\n",
			      fd, fd_lifetime[fd], cached_curtime);
			comm_close(fd);
		    } else {
			/* seems like handle closed it. 
			   clean up fd_table just to make sure */
			debug(5, "comm_select - %d lifetime expire: %d : Handler closed the socket.\n",
			      fd, fd_lifetime[fd]);
			/* just to make sure here */
			comm_cleanup_fd_entry(fd);
		    }
		} else {
		    /* no handle. do full comm_close */
		    debug(5, "comm_select - %d lifetime expire: %d < %d : No handler to close the socket.\n comm_select will do.\n",
			  fd, fd_lifetime[fd], cached_curtime);
		    comm_close(fd);
		}
	    }
	}

	/* scan each socket */
	if (num) {
	    for (fd = 0; (fd < (fdstat_biggest_fd() + 1)) && (num > 0); fd++) {
		if (FD_ISSET(fd, &readfds)) {
		    debug(6, "comm_select - %d ready for reading\n", fd);
		    --num;
		    if (fd_table[fd].read_handler) {
			int (*tmp) () = fd_table[fd].read_handler;
			fd_table[fd].read_handler = 0;
			debug(10, "calling read handler %p(%d,%p)\n",
			      tmp, fd, fd_table[fd].read_data);
			tmp(fd, fd_table[fd].read_data);
		    }
		}
		if (FD_ISSET(fd, &writefds)) {
		    debug(5, "comm_select - %d ready for writing\n", fd);
		    --num;
		    if (fd_table[fd].write_handler) {
			int (*tmp) () = fd_table[fd].write_handler;
			fd_table[fd].write_handler = 0;
			debug(10, "calling write handler %p(%d,%p)\n",
			      tmp, fd, fd_table[fd].write_data);
			tmp(fd, fd_table[fd].write_data);
		    }
		}
		if (FD_ISSET(fd, &exceptfds)) {
		    --num;
		    debug(5, "comm_select - %d has an exception\n", fd);
		    if (fd_table[fd].except_handler) {
			int (*tmp) () = fd_table[fd].except_handler;
			fd_table[fd].except_handler = 0;
			debug(10, "calling except handler %p(%d,%p)\n",
			      tmp, fd, fd_table[fd].except_data);
			tmp(fd, fd_table[fd].except_data);
		    }
		}
	    }
	    return COMM_OK;
	}
    }

    debug(8, "comm_select: time out: %d.\n", cached_curtime);
    return COMM_TIMEOUT;
}


/* Select on fd to see if any io pending. */
int comm_pending(fd, sec, usec)
int fd;
long sec, usec;
{
    fd_set readfds;
    int num;
    struct timeval timeout;

    /* Find a fd ready for reading. */
    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);
    timeout.tv_sec = sec;
    timeout.tv_usec = usec;

    while ((num = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout)) < 0) {
	if (errno == EWOULDBLOCK)
	    return COMM_NOMESSAGE;
	else if (errno != EINTR) {
	    debug(1, "comm_pending: select failure: %s (errno %d).\n",
		  sys_errlist[errno], errno);
	    return COMM_ERROR;
	}
	/* if select interrupted, try again */
    }

    debug(5, "comm_pending - %d sockets ready for reading\n", num);

    if (num && FD_ISSET(fd, &readfds)) {
	return COMM_OK;
    }
    return COMM_TIMEOUT;
}

int comm_set_select_handler(fd, type, handler, client_data)
int fd;
unsigned int type;
/* 01 - read; 10 - write; 100 - except; 1000 - timeout ; 10000 - lifetime */
int (*handler) ();
caddr_t client_data;
{

    return (comm_set_select_handler_plus_timeout(fd, type, handler, client_data, 0));
}

/* Should use var args here PBD */
int comm_set_select_handler_plus_timeout(fd, type, handler, client_data, timeout)
int fd;
unsigned int type;
/* 01 - read; 10 - write; 100 - except; 1000 - timeout ; 10000 - lifetime */
int (*handler) ();
caddr_t client_data;
time_t timeout;
{
    if (type & COMM_SELECT_TIMEOUT) {
	fd_table[fd].timeout_time = (time(0L) + timeout);
	fd_table[fd].timeout_delta = timeout;
	fd_table[fd].timeout_handler = handler;
	fd_table[fd].timeout_data = client_data;
	if ((timeout <= 0) && handler) {
	    debug(2, "comm_set_select_handler_plus_timeout: Zero timeout doesn't make sense\n");
	}
    }
    if (type & COMM_SELECT_READ) {
	fd_table[fd].read_handler = handler;
	fd_table[fd].read_data = client_data;
    }
    if (type & COMM_SELECT_WRITE) {
	fd_table[fd].write_handler = handler;
	fd_table[fd].write_data = client_data;
    }
    if (type & COMM_SELECT_EXCEPT) {
	fd_table[fd].except_handler = handler;
	fd_table[fd].except_data = client_data;
    }
    if (type & COMM_SELECT_LIFETIME) {
	fd_table[fd].lifetime_handler = handler;
	fd_table[fd].lifetime_data = client_data;
    }
    return 0;			/* XXX What is meaningful? */
}

int comm_get_select_handler(fd, type, handler_ptr, client_data_ptr)
int fd;
unsigned int type;
int (**handler_ptr) ();
caddr_t *client_data_ptr;
{
    if (type & COMM_SELECT_TIMEOUT) {
	*handler_ptr = fd_table[fd].timeout_handler;
	*client_data_ptr = fd_table[fd].timeout_data;
    }
    if (type & COMM_SELECT_READ) {
	*handler_ptr = fd_table[fd].read_handler;
	*client_data_ptr = fd_table[fd].read_data;
    }
    if (type & COMM_SELECT_WRITE) {
	*handler_ptr = fd_table[fd].write_handler;
	*client_data_ptr = fd_table[fd].write_data;
    }
    if (type & COMM_SELECT_EXCEPT) {
	*handler_ptr = fd_table[fd].except_handler;
	*client_data_ptr = fd_table[fd].except_data;
    }
    if (type & COMM_SELECT_LIFETIME) {
	*handler_ptr = fd_table[fd].lifetime_handler;
	*client_data_ptr = fd_table[fd].lifetime_data;
    }
    return 0;			/* XXX What is meaningful? */
}


int setsocketNolinger(fd)
int fd;
{
    int off = 0;

    debug(10, "setsocketNolinger: turning off SO_LINGER on FD %d\n", fd);
    return setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *) &off, sizeof(off));
}

int setsocketReuse(fd)
int fd;
{
    int on = 1;

    debug(10, "setsocketReuse: turning on SO_REUSEADDR on FD %d\n", fd);
    return setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));
}

int setNonBlocking(new_socket)
int new_socket;
{
    debug(10, "setNonBlocking: setting FD %d to non-blocking i/o.\n",
	  new_socket);
    /*
     * Set up the flag NOT to have the socket to wait for message from
     * network forever, but to return -1 when no message is coming in.
     */

#ifdef __hpux
    if (fcntl(new_socket, F_SETFL, O_NONBLOCK)) {
	debug(0, "comm_open: error setting O_NONBLOCK: %s\n",
	      sys_errlist[errno]);
	return (COMM_ERROR);
    }
#else
    if (fcntl(new_socket, F_SETFL, FNDELAY)) {
	debug(0, "comm_open: error setting FNDELAY: %s\n",
	      sys_errlist[errno]);
	return (COMM_ERROR);
    }
#endif				/* HPUX */
    return 0;
}


struct in_addr *
 getAddress(name)
char *name;
{

    char hostname[256];
    struct hostent *hp;
    struct in_addr *our_first_addr;

    if (name == NULL) {
	return NULL;
    } else {
	strcpy(hostname, name);
    }

    if ((hp = ipcache_gethostbyname(hostname)) == 0) {
	debug(0, "getAddress: gethostbyname failure: %s\n",
	      sys_errlist[errno]);
	return (NULL);
    } else {
	our_first_addr = (struct in_addr *) hp->h_addr;
	return (our_first_addr);
    }
}

/*
 *  the fd_lifetime is used as a hardlimit to timeout dead sockets.
 *  The basic problem is that many WWW clients are abusive and
 *  it results in cached having lots of CLOSE_WAIT states.  Until
 *  we can find a better solution, we give all asciiPort or
 *  cached initiated clients a maximum lifetime.
 */
int comm_init()
{
    int i, max_fd = getMaxFD();

    fd_table = (FD_ENTRY *) xmalloc(sizeof(FD_ENTRY) * max_fd);
    memset(fd_table, '\0', sizeof(FD_ENTRY) * max_fd);	/* clear fd_table */

    /* hardwired lifetimes */
    fd_lifetime = (int *) xmalloc(sizeof(int) * max_fd);
    for (i = 0; i < max_fd; i++) {
	fd_lifetime[i] = -1;	/* denotes invalid */
    }
    return 0;
}


/*
 * examine_select - debug routine.
 *
 * I spend the day chasing this core dump that occurs when both the client
 * and the server side of a cache fetch simultaneoulsy abort the
 * connection.  While I haven't really studied the code to figure out how
 * it happens, the snippet below may prevent the cache from exitting:
 * 
 * Call this from where the select loop fails.
 */
int examine_select(readfds, writefds, exceptfds)
fd_set *readfds, *writefds, *exceptfds;
{
    int fd = 0;
    fd_set read_x, write_x, except_x;
    int num;
    struct timeval tv;

    debug(0, "examine_select: Examining open file descriptors...\n");
    tv.tv_sec = tv.tv_usec = 0;
    for (fd = 0; fd < getMaxFD(); ++fd) {
	FD_ZERO(&read_x);
	FD_ZERO(&write_x);
	FD_ZERO(&except_x);
	if ((FD_ISSET(fd, readfds)) ||
	    (FD_ISSET(fd, writefds)) ||
	    (FD_ISSET(fd, exceptfds))) {
	    FD_SET(fd, &read_x);
	    num = select(FD_SETSIZE, &read_x, &read_x, &read_x, &tv);
	    if (num < 0) {
		debug(0, "WARNING: FD %d has handlers, but it's invalid.\n", fd);
		debug(0, "Timeout handler:%x read:%x write:%x except:%x\n",
		      fd_table[fd].timeout_handler,
		      fd_table[fd].read_handler,
		      fd_table[fd].write_handler,
		      fd_table[fd].except_handler);
		fd_table[fd].timeout_handler = 0;
		fd_table[fd].read_handler = 0;
		fd_table[fd].write_handler = 0;
		fd_table[fd].except_handler = 0;
		FD_CLR(fd, readfds);
		FD_CLR(fd, writefds);
		FD_CLR(fd, exceptfds);
	    }
	}
    }
    debug(0, "examine_select: Finished examining open file descriptors.\n");
    return 0;
}
