static char rcsid[] = "$Id: dnsserver.c,v 1.22 1995/08/04 04:01:20 hardy Exp $";
/*
 *  File:         dnsserver.c
 *  Description:  dnsserver process for non-blocking DNS lookup.
 *  Author:       Anawat Chankhunthod
 *  Created:
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder.
 *          Duane Wessels of the University of Colorado at Boulder.
 *
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *
 *  - You must include the above copyright notice and these accompanying
 *    paragraphs in all forms of derivative works, and any documentation
 *    and other materials related to such distribution and use acknowledge
 *    that the software was developed at the above institutions.
 *
 *  - You must notify IRTF-RD regarding your distribution of the
 *    derivative work.
 *
 *  - You must clearly notify users that your are distributing a modified
 *    version and not the original Harvest software.
 *
 *  - Any derivative product is also subject to the restrictions of the
 *    copyright, including distribution and use limitations.
 *
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>

#include "config.h"

extern int h_errno;

int debug_flag = 0;

/* error messages from gethostbyname() */
#define my_h_msgs(x) (\
	((x) == HOST_NOT_FOUND) ? \
		"Authoritative Answer Host not found" : \
	((x) == TRY_AGAIN) ? \
		"Non-Authoritive Host not found" : \
	((x) == NO_RECOVERY) ? \
		"Non recoverable errors" : \
	((x) == NO_DATA) ? \
		"Valid name, no data record of requested type" : \
	((x) == NO_ADDRESS) ? \
		"No address, look for MX record" : \
		"Unknown DNS problem?")

int main(argc, argv)
int argc;
char *argv[];
{
    char request[256], msg[256];
    struct hostent *result = NULL;
    FILE *logfile = NULL;
    long start;
    long stop;
    char *t = NULL;
    char buf[256];

    while (argc > 1) {
	argc--;
	if ((!strcmp(argv[argc], "-v"))
	    || (!strcmp(argv[argc], "-h"))
	    || (!strcmp(argv[argc], "-help"))) {
	    printf("dnsserver version %s\n", HARVEST_VERSION);
	    exit(0);
	} else if (!strcmp(argv[argc], "-d")) {
	    sprintf(buf, "dnsserver.%d.log", getpid());
	    logfile = fopen(buf, "a");
	    debug_flag++;
	    if (!logfile)
		fprintf(stderr, "Could not open dnsserver's log file\n");
	}
    }

    while (1) {
	memset(request, '\0', 256);
	if (fgets(request, 255, stdin) == (char *) NULL)
	    exit(1);
	if ((t = strrchr(request, '\n')) != NULL)
	    *t = '\0';		/* strip NL */
	if ((t = strrchr(request, '\r')) != NULL)
	    *t = '\0';		/* strip CR */
	if (strcmp(request, "$shutdown") == 0)
	    exit(0);

	if (strcmp(request, "$hello") == 0) {
	    printf("$alive\n");
	    printf("$end\n");
	    fflush(stdout);
	    continue;
	}
	/* check if it's already a IP address in text form. */
	if (isdigit(request[0])) {
	    int a1, a2, a3, a4;
	    if (sscanf(request, "%u.%u.%u.%u", &a1, &a2, &a3, &a4) != 4) {
		printf("$fail %s\n", request);
		printf("$message Not a valid IP number\n");
		printf("$end\n");
		fflush(stdout);
		continue;
	    } else {
		printf("$name %s\n", request);
		printf("$h_name %s\n", request);
		printf("$h_len %d\n", 4);
		printf("$ipcount %d\n", 1);
		printf("%s\n", request);
		printf("$aliascount %d\n", 0);
		printf("$end\n");
		fflush(stdout);
		continue;
	    }
	}
	start = time(NULL);
	result = gethostbyname(request);
	if (!result) {
	    if (h_errno == TRY_AGAIN) {
		sleep(2);
		result = gethostbyname(request);	/* try a little harder */
	    }
	}
	stop = time(NULL);

	msg[0] = '\0';
	if (!result) {
	    if (h_errno == TRY_AGAIN) {
		sprintf(msg, "Name Server for domain '%s' is unavailable.",
			request);
	    } else {
		sprintf(msg, "DNS Domain '%s' is invalid: %s.\n",
			request, my_h_msgs(h_errno));
	    }
	}
	if (!result || (strlen(result->h_name) == 0)) {
	    if (logfile) {
		fprintf(logfile, "%s %d\n", request, (int) (stop - start));
		fflush(logfile);
	    }
	    printf("$fail %s\n", request);
	    printf("$message %s\n", msg[0] ? msg : "Unknown Error");
	    printf("$end\n");
	    fflush(stdout);
	    continue;
	} else {
	    int addr_count = 0;
	    int alias_count = 0;
	    int i;

	    printf("$name %s\n", request);
	    printf("$h_name %s\n", result->h_name);
	    printf("$h_len %d\n", result->h_length);

	    while ((addr_count < 255) && result->h_addr_list[addr_count])
		++addr_count;
	    printf("$ipcount %d\n", addr_count);
	    for (i = 0; i < addr_count; i++) {
		struct in_addr addr;
		memcpy((char *) &addr, result->h_addr_list[i], result->h_length);
		printf("%s\n", inet_ntoa(addr));
	    }

	    while ((alias_count < 255) && result->h_aliases[alias_count])
		++alias_count;
	    printf("$aliascount %d\n", alias_count);
	    for (i = 0; i < alias_count; i++) {
		printf("%s\n", result->h_aliases[i]);
	    }

	    printf("$end\n");
	    fflush(stdout);
	    continue;
	}
    }

    exit(0);
}
