static char rcsid[] = "$Id: gopher.c,v 1.52 1995/08/21 21:23:15 duane Exp $";
/* 
 * File:         gopher.c
 * Description:  state machine for gopher retrieval protocol.  Based on Anawat's
 *               gopher retrieval module.
 * Author:       John Noll, USC, Anawat
 * Created:      Thu Apr 28 10:57:11 1994
 * Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <sys/param.h>		/* For MAXHOSTNAMELEN. */
#include <sys/errno.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "comm.h"
#include "store.h"
#include "stat.h"
#include "url.h"
#include "mime.h"
#include "cache_cf.h"
#include "ttl.h"
#include "config.h"

extern int errno;
extern char *sys_errlist[];
extern char *dns_error_message;

/* gopher type code from rfc. Anawat. */
#define GOPHER_FILE         '0'
#define GOPHER_DIRECTORY    '1'
#define GOPHER_CSO          '2'
#define GOPHER_ERROR        '3'
#define GOPHER_MACBINHEX    '4'
#define GOPHER_DOSBIN       '5'
#define GOPHER_UUENCODED    '6'
#define GOPHER_INDEX        '7'
#define GOPHER_TELNET       '8'
#define GOPHER_BIN          '9'
#define GOPHER_REDUNT       '+'
#define GOPHER_3270         'T'
#define GOPHER_GIF          'g'
#define GOPHER_IMAGE        'I'

#define GOPHER_HTML         'h'	/* HTML */
#define GOPHER_INFO         'i'
#define GOPHER_WWW          'w'	/* W3 address */
#define GOPHER_SOUND        's'

#define GOPHER_PLUS_IMAGE   ':'
#define GOPHER_PLUS_MOVIE   ';'
#define GOPHER_PLUS_SOUND   '<'

#define GOPHER_PORT         70

#define TAB                 '\t'
#define TEMP_BUF_SIZE       4096
#define MAX_CSO_RESULT      1024

typedef struct gopher_ds {
    StoreEntry *entry;
    char host[MAXHOSTNAMELEN];
    enum {
	NORMAL,
	HTML_DIR,
	HTML_INDEX_RESULT,
	HTML_CSO_RESULT,
	HTML_INDEX_PAGE,
	HTML_CSO_PAGE
    } conversion;
    int HTML_header_added;
    int port;
    char *mime_hdr;
    char type_id;
    char *tail;
    char request[MAX_URL];
    int data_in;
    int cso_recno;
    int len;
    char buf[TEMP_BUF_SIZE];
} GopherData;

extern char *tmp_error_buf;

#define       gopherCASE(t,s) \
		case t: strcat(tempMIME, s); break;
#define       gopherCASE2(t1,t2,s) \
		case t1: case t2: strcat(tempMIME, s); break;
#define       gopherCASE3(t1,t2,t3,s) \
		case t1: case t2: case t3: strcat(tempMIME, s); break;
#define       gopherCASE4(t1,t2,t3,t4,s) \
		case t1: case t2: case t3: case t4: \
                strcat(tempMIME, s); break;
#define       gopherCASE5(t1,t2,t3,t4,t5,s) \
		case t1: case t2: case t3: case t4: \
		case t5: strcat(tempMIME, s); break;
char def_gopher_bin[] = "www/unknown";
char def_gopher_text[] = "text/plain";

/* figure out content type from file extension */
static void gopher_mime_content(buf, name, def)
char *buf;
char *name;
char *def;
{
    char *ext1, *ext2, temp[MAX_URL], *str;
    ext_table_entry *e;

    ext2 = NULL;
    strcpy(temp, name);
    for (ext1 = temp; *ext1; ext1++)
	if (isupper(*ext1))
	    *ext1 = tolower(*ext1);
    if ((ext1 = strrchr(temp, '.')) == NULL) {
	/* use default */
	sprintf(buf + strlen(buf), "Content-Type: %s\r\n", def);
	return;
    }
    /* try extension table */
    *ext1++ = 0;
    if (strcmp("gz", ext1) == 0 || strcmp("z", ext1) == 0) {
	ext2 = ext1;
	if ((ext1 = strrchr(temp, '.')) == NULL) {
	    ext1 = ext2;
	    ext2 = NULL;
	} else
	    ext1++;
    }
    if ((e = mime_ext_to_type(ext1)) == NULL) {
	if (ext2) {
	    e = mime_ext_to_type(ext2);
	    str = e->mime_type;
	    ext2 = NULL;
	} else
	    str = def;
    } else
	str = e->mime_type;
    sprintf(buf + strlen(buf), "Content-Type: %s\r\n", str);
    if (ext2) {
	e = mime_ext_to_type(ext2);
	sprintf(buf + strlen(buf), "Content-Encoding: %s\r\n",
		e->mime_encoding);
    }
}



/* create MIME Header for Gopher Data */
void gopherMimeCreate(data)
GopherData *data;
{
    char tempMIME[MAX_MIME];

    sprintf(tempMIME, "\
HTTP/1.0 200 OK Gatewaying\r\n\
Server: HarvestCache/%s\r\n\
MIME-version: 1.0\r\n", HARVEST_VERSION);

    switch (data->type_id) {

	gopherCASE5(GOPHER_DIRECTORY, GOPHER_INDEX, GOPHER_HTML, GOPHER_WWW,
		    GOPHER_CSO, "Content-Type: text/html\r\n");
	gopherCASE(GOPHER_GIF, "Content-Type: image/gif\r\n");
	gopherCASE2(GOPHER_SOUND, GOPHER_PLUS_SOUND, "Content-Type: audio/void\r\n");
	gopherCASE2(GOPHER_IMAGE, GOPHER_PLUS_IMAGE, "Content-Type: image/void\r\n");
	gopherCASE(GOPHER_PLUS_MOVIE, "Content-Type: video/void\r\n");

    case GOPHER_MACBINHEX:
    case GOPHER_DOSBIN:
    case GOPHER_UUENCODED:
    case GOPHER_BIN:
	/* Rightnow We have no idea what it is. */
	gopher_mime_content(tempMIME, data->request, def_gopher_bin);
	break;

    case GOPHER_FILE:
    default:
	gopher_mime_content(tempMIME, data->request, def_gopher_text);
	break;

    }

    strcat(tempMIME, "\r\n\r\n");
    storeAppend(data->entry, tempMIME, strlen(tempMIME));
}

/* Parse a gopher url into components.  By Anawat. */
int gopher_url_parser(url, host, port, type_id, request)
char *url;
char *host;
int *port;
char *type_id;
char *request;
{
    char atypebuf[MAX_URL], hostbuf[MAX_URL], *tmp;
    int t;

    atypebuf[0] = hostbuf[0] = '\0';
    host[0] = request[0] = '\0';
    (*port) = 0;
    (*type_id) = 0;

    t = sscanf(url, "%[a-zA-Z]://%[^/]/%c%s", atypebuf, hostbuf,
	       type_id, request);
    if ((t < 2) || strcasecmp(atypebuf, "gopher")) {
	return -1;
    } else if (t == 2) {
	(*type_id) = GOPHER_DIRECTORY;
	request[0] = '\0';
    } else if (t == 3) {
	request[0] = '\0';
    } else {
	/* convert %xx to char */
	tmp = url_convert_hex(request);
	strncpy(request, tmp, MAX_URL);
	safe_free(tmp);
    }

    host[0] = '\0';
    if (sscanf(hostbuf, "%[^:]:%d", host, port) < 2)
	(*port) = GOPHER_PORT;

    return 0;
}

int gopherCachable(url, type, mime_hdr)
char *url;
char *type;
char *mime_hdr;
{
    stoplist *p;

    /* scan stop list */
    p = gopher_stoplist;
    while (p) {
	if (strstr(url, p->key))
	    return 0;
	p = p->next;
    }

    {
	/* use as temp data structure to parse gopher URL */
	GopherData *data = xcalloc(1, sizeof(GopherData));

	/* parse to see type */
	gopher_url_parser(url, data->host, &data->port, &data->type_id, data->request);

	if ((data->type_id == GOPHER_INDEX) ||
	    (data->type_id == GOPHER_CSO) ||
	    (data->type_id == GOPHER_TELNET) ||
	    (data->type_id == GOPHER_3270)) {
	    safe_free(data);
	    return 0;
	}
	safe_free(data);
    }

    /* else cachable */
    return 1;

}

void gopherEndHTML(data)
GopherData *data;
{
    char tmpbuf[TEMP_BUF_SIZE];

    if (!data->data_in) {
	sprintf(tmpbuf, "<HR><H2><i>Server Return Nothing.</i></H2>\n");
	storeAppend(data->entry, tmpbuf, strlen(tmpbuf));
	return;
    }
}


/* Convert Gopher to HTML */
/* Borrow part of code from libwww2 came with Mosaic distribution */
void gopherToHTML(data, inbuf, len)
GopherData *data;
char *inbuf;
int len;
{
    char *pos = inbuf;
    char *lpos;
    char *tline;
    char *line = xcalloc(1, TEMP_BUF_SIZE);
    char *outbuf = xcalloc(1, TEMP_BUF_SIZE * 16);
    char *tmpbuf = xcalloc(1, TEMP_BUF_SIZE);
    char *name, *selector, *host, *port, *escaped_selector;
    char *icon_type;
    char gtype;


    if (data->conversion == HTML_INDEX_PAGE) {
	sprintf(outbuf, "<TITLE>Gopher Index %s</TITLE><H1>%s<BR>Gopher Search</H1> This is a searchable Gopher index.Use the search function of your browser to enter search terms. <ISINDEX>\n", data->entry->url, data->entry->url);

	storeAppend(data->entry, outbuf, strlen(outbuf));
	/* now let start sending stuff to client */
	BIT_RESET(data->entry->flag, DELAY_SENDING);
	data->data_in = 1;

	safe_free(line);
	safe_free(outbuf);
	safe_free(tmpbuf);
	return;
    }
    if (data->conversion == HTML_CSO_PAGE) {
	sprintf(outbuf, "<TITLE>CSO Search of %s</TITLE><H1>%s<BR>CSO Search</H1>A CSO database usually contains a phonebook or directory. Use the search function of your browser to enter search terms.<ISINDEX>\n",
		data->entry->url, data->entry->url);

	storeAppend(data->entry, outbuf, strlen(outbuf));
	/* now let start sending stuff to client */
	BIT_RESET(data->entry->flag, DELAY_SENDING);
	data->data_in = 1;

	safe_free(line);
	safe_free(outbuf);
	safe_free(tmpbuf);
	return;
    }
    inbuf[len] = '\0';

    if (!data->HTML_header_added) {
	if (data->conversion == HTML_CSO_RESULT)
	    strcat(outbuf, "<H1>CSO Searchs Result</H1>\n<PRE>\n");
	else
	    strcat(outbuf, "<H1>Gopher Menu</H1>\n<PRE>\n");
	data->HTML_header_added = 1;
    }
    while (pos < inbuf + len) {

	if (data->len != 0) {
	    /* there is something left from last tx. */
	    strncpy(line, data->buf, data->len);
	    lpos = (char *) memccpy(line + data->len, inbuf, '\n', len);
	    if (lpos)
		*lpos = '\0';
	    else {
		/* there is no complete line in inbuf */
		/* copy it to temp buffer */
		if (data->len + len > TEMP_BUF_SIZE) {
		    debug(1, "GopherHTML: Buffer overflow. Lost some data on URL: %s\n",
			  data->entry->url);
		    len = TEMP_BUF_SIZE - data->len;
		}
		memcpy(data->buf + data->len, inbuf, len);
		data->len += len;
		safe_free(line);
		safe_free(outbuf);
		safe_free(tmpbuf);
		return;
	    }

	    /* skip one line */
	    pos = (char *) memchr(pos, '\n', 256);
	    if (pos)
		pos++;

	    /* we're done with the remain from last tx. */
	    data->tail = data->buf;
	    data->len = 0;
	    *(data->buf) = '\0';
	} else {

	    lpos = (char *) memccpy(line, pos, '\n', len - (pos - inbuf));
	    if (lpos)
		*lpos = '\0';
	    else {
		/* there is no complete line in inbuf */
		/* copy it to temp buffer */
		if ((len - (pos - inbuf)) > TEMP_BUF_SIZE) {
		    debug(1, "GopherHTML: Buffer overflow. Lost some data on URL: %s\n",
			  data->entry->url);
		    len = TEMP_BUF_SIZE;
		}
		if (len > (pos - inbuf)) {
		    memcpy(data->buf, pos, len - (pos - inbuf));
		    data->len = len - (pos - inbuf);
		}
		break;
	    }

	    /* skip one line */
	    pos = (char *) memchr(pos, '\n', 256);
	    if (pos)
		pos++;

	}

	/* at this point. We should have one line in buffer to process */

	if (*line == '.') {
	    /* skip it */
	    memset(line, '\0', TEMP_BUF_SIZE);
	    continue;
	}
	switch (data->conversion) {

	case HTML_INDEX_RESULT:
	case HTML_DIR:{
		tline = line;
		gtype = *tline++;
		name = tline;
		selector = strchr(tline, TAB);
		if (selector) {
		    *selector++ = '\0';
		    host = strchr(selector, TAB);
		    if (host) {
			*host++ = '\0';
			port = strchr(host, TAB);
			if (port) {
			    char *junk;
			    port[0] = ':';
			    junk = strchr(host, TAB);
			    if (junk)
				*junk++ = 0;	/* Chop port */
			    else {
				junk = strchr(host, '\r');
				if (junk)
				    *junk++ = 0;	/* Chop port */
				else {
				    junk = strchr(host, '\n');
				    if (junk)
					*junk++ = 0;	/* Chop port */
				}
			    }
			    if ((port[1] == '0') && (!port[2]))
				port[0] = 0;	/* 0 means none */
			}
			/* escape a selector here */
			escaped_selector = url_escape(selector);

			switch (gtype) {
			case GOPHER_DIRECTORY:
			    icon_type = "internal-gopher-menu";
			    break;
			case GOPHER_FILE:
			    icon_type = "internal-gopher-text";
			    break;
			case GOPHER_INDEX:
			case GOPHER_CSO:
			    icon_type = "internal-gopher-index";
			    break;
			case GOPHER_IMAGE:
			case GOPHER_GIF:
			case GOPHER_PLUS_IMAGE:
			    icon_type = "internal-gopher-image";
			    break;
			case GOPHER_SOUND:
			case GOPHER_PLUS_SOUND:
			    icon_type = "internal-gopher-sound";
			    break;
			case GOPHER_PLUS_MOVIE:
			    icon_type = "internal-gopher-movie";
			    break;
			case GOPHER_TELNET:
			case GOPHER_3270:
			    icon_type = "internal-gopher-telnet";
			    break;
			case GOPHER_BIN:
			case GOPHER_MACBINHEX:
			case GOPHER_DOSBIN:
			case GOPHER_UUENCODED:
			    icon_type = "internal-gopher-binary";
			    break;
			default:
			    icon_type = "internal-gopher-unknown";
			    break;
			}


			memset(tmpbuf, '\0', TEMP_BUF_SIZE);
			if ((gtype == GOPHER_TELNET) || (gtype == GOPHER_3270)) {
			    if (strlen(escaped_selector) != 0)
				sprintf(tmpbuf, "<IMG BORDER=0 SRC=\"%s\"> <A HREF=\"telnet://%s@%s/\">%s</A>\n",
				icon_type, escaped_selector, host, name);
			    else
				sprintf(tmpbuf, "<IMG BORDER=0 SRC=\"%s\"> <A HREF=\"telnet://%s/\">%s</A>\n",
					icon_type, host, name);

			} else {
			    sprintf(tmpbuf, "<IMG BORDER=0 SRC=\"%s\"> <A HREF=\"gopher://%s/%c%s\">%s</A>\n",
				    icon_type, host, gtype, escaped_selector, name);
			}
			safe_free(escaped_selector);
			strcat(outbuf, tmpbuf);
			data->data_in = 1;
		    } else {
			memset(line, '\0', TEMP_BUF_SIZE);
			continue;
		    }
		} else {
		    memset(line, '\0', TEMP_BUF_SIZE);
		    continue;
		}
		break;
	    }			/* HTML_DIR, HTML_INDEX_RESULT */


	case HTML_CSO_RESULT:{
		int t;
		int code;
		int recno;
		char result[MAX_CSO_RESULT];

		tline = line;

		if (tline[0] == '-') {
		    t = sscanf(tline, "-%d:%d:%[^\n]", &code, &recno, result);
		    if (t < 3)
			break;

		    if (code != 200)
			break;

		    if (data->cso_recno != recno) {
			sprintf(tmpbuf, "</PRE><HR><H2>Record# %d<br><i>%s</i></H2>\n<PRE>", recno, result);
			data->cso_recno = recno;
		    } else {
			sprintf(tmpbuf, "%s\n", result);
		    }
		    strcat(outbuf, tmpbuf);
		    data->data_in = 1;
		    break;
		} else {
		    /* handle some error codes */
		    t = sscanf(tline, "%d:%[^\n]", &code, result);

		    if (t < 2)
			break;

		    switch (code) {

		    case 200:{
			    /* OK */
			    /* Do nothing here */
			    break;
			}

		    case 102:	/* Number of matches */
		    case 501:	/* No Match */
		    case 502:	/* Too Many Matches */
			{
			    /* Print the message the server returns */
			    sprintf(tmpbuf, "</PRE><HR><H2><i>%s</i></H2>\n<PRE>", result);
			    strcat(outbuf, tmpbuf);
			    data->data_in = 1;
			    break;
			}


		    }
		}

	    }			/* HTML_CSO_RESULT */
	default:
	    break;		/* do nothing */

	}			/* switch */

    }				/* while loop */

    if (strlen(outbuf) > 0) {
	storeAppend(data->entry, outbuf, strlen(outbuf));
	/* now let start sending stuff to client */
	BIT_RESET(data->entry->flag, DELAY_SENDING);
    }
    safe_free(line);
    safe_free(outbuf);
    safe_free(tmpbuf);
    return;
}


int gopherReadReplyTimeout(fd, data)
int fd;
GopherData *data;
{
    debug(4, "GopherReadReplyTimeout: Timeout on %d\n url: %s\n", fd, data->entry->url);
    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
	    data->entry->url,
	    data->entry->url,
	    "Gopher",
	    203,
	    "Read timeout",
	    "Network/Remote site may be down. Try again later",
	    HARVEST_VERSION,
	    comm_hostname());
    storeAbort(data->entry, tmp_error_buf);
    comm_close(fd);
    safe_free(data);
    return 0;
}

/* This will be called when socket lifetime is expired. */
void gopherLifetimeExpire(fd, data)
int fd;
GopherData *data;
{
    debug(4, "gopherLifeTimeExpire: FD %d: <URL:%s>\n", fd, data->entry->url);
    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
	    data->entry->url,
	    data->entry->url,
	    "GOPHER",
	    110,
	    "Transaction Timeout",
      "The Network/Remote site may be down or too slow. Try again later",
	    HARVEST_VERSION,
	    comm_hostname());
    storeAbort(data->entry, tmp_error_buf);
    comm_set_select_handler(fd, COMM_SELECT_READ, 0, 0);
    comm_close(fd);
    safe_free(data);
}




/* This will be called when data is ready to be read from fd.  Read until
   error or connection closed. */
int gopherReadReply(fd, data)
int fd;
GopherData *data;
{
    char buf[4096];
    int len = read(fd, buf, 4096);
    debug(5, "gopherReadReply - fd: %d read len:%d\n", fd, len);

    if (len < 0) {
	debug(1, "gopherReadReply - error reading: %s\n",
	      sys_errlist[errno]);
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"Gopher",
		205,
		"Read error",
		"Network/Remote Site is down.  Try again later.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(data->entry, tmp_error_buf);
	comm_close(fd);
	safe_free(data);

    } else if (len == 0) {
	/* Connection closed; retrieval done. */
	/* flush the rest of data in temp buf if there is one. */
	if (data->conversion != NORMAL)
	    gopherEndHTML(data);
	data->entry->ttl = ttlSet(data->entry);
	BIT_RESET(data->entry->flag, DELAY_SENDING);
	storeComplete(data->entry);
	comm_close(fd);
	safe_free(data);

    } else if ((data->entry->e_current_len + len) > getGopherMax()) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"Gopher",
		206,
		"Object is too large",
		"Contact your system manager for further help, or retrieve the object directly without a proxy.",
		HARVEST_VERSION,
		comm_hostname());
	if (data->conversion != NORMAL)
	    gopherEndHTML(data);
	BIT_RESET(data->entry->flag, DELAY_SENDING);
	storeAbort(data->entry, tmp_error_buf);
	comm_close(fd);
	safe_free(data);
    } else if (data->entry->flag & CLIENT_ABORT_REQUEST) {
	/* append the last bit of info we got */
	if (data->conversion != NORMAL) {
	    gopherToHTML(data, buf, len);
	} else {
	    storeAppend(data->entry, buf, len);
	}
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"Gopher",
		207,
		"Client Aborted",
		"Client(s) dropped connection before transmission is complete.\nObject fetching is aborted.\n",
		HARVEST_VERSION,
		comm_hostname());
	if (data->conversion != NORMAL)
	    gopherEndHTML(data);
	BIT_RESET(data->entry->flag, DELAY_SENDING);
	storeAbort(data->entry, tmp_error_buf);
	comm_close(fd);
	safe_free(data);

    } else {
	if (data->conversion != NORMAL) {
	    gopherToHTML(data, buf, len);
	} else {
	    storeAppend(data->entry, buf, len);
	}
	comm_set_select_handler(fd, COMM_SELECT_READ, (PF) gopherReadReply, (caddr_t) data);
	comm_set_select_handler_plus_timeout(fd, COMM_SELECT_TIMEOUT, (PF) gopherReadReplyTimeout,
				       (caddr_t) data, getReadTimeout());
    }
    return 0;
}

/* This will be called when request write is complete. Schedule read of
   reply. */
int gopherSendComplete(fd, buf, size, errflag, data)
int fd;
char *buf;
int size;
int errflag;
GopherData *data;
{
    debug(5, "gopherSendComplete - fd: %d size: %d errflag: %d\n",
	  fd, size, errflag);
    if (errflag) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		data->entry->url,
		data->entry->url,
		"Gopher",
		201,
		"Cannot connect to the original site",
		"The remote site may be down.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(data->entry, tmp_error_buf);
	comm_close(fd);
	safe_free(data);
    } else {
	/* OK. We successfully reach remote site. */
	/* start MIME typing stuff. Do it anyway even though request is not HTML type. */
	gopherMimeCreate(data);


	if (!BIT_TEST(data->entry->flag, REQ_HTML))
	    data->conversion = NORMAL;
	else {
	    switch (data->type_id) {

	    case GOPHER_DIRECTORY:{
		    /* we got to convert it first */
		    BIT_SET(data->entry->flag, DELAY_SENDING);
		    data->conversion = HTML_DIR;
		    data->HTML_header_added = 0;
		    break;
		}

	    case GOPHER_INDEX:{
		    /* we got to convert it first */
		    BIT_SET(data->entry->flag, DELAY_SENDING);
		    data->conversion = HTML_INDEX_RESULT;
		    data->HTML_header_added = 0;
		    break;
		}

	    case GOPHER_CSO:{
		    /* we got to convert it first */
		    BIT_SET(data->entry->flag, DELAY_SENDING);
		    data->conversion = HTML_CSO_RESULT;
		    data->cso_recno = 0;
		    data->HTML_header_added = 0;
		    break;
		}


	    default:{
		    data->conversion = NORMAL;
		}

	    }
	}
	/* Schedule read reply. */
	comm_set_select_handler(fd, COMM_SELECT_READ, (PF) gopherReadReply, (caddr_t) data);
	comm_set_select_handler_plus_timeout(fd, COMM_SELECT_TIMEOUT, (PF) gopherReadReplyTimeout,
				       (caddr_t) data, getReadTimeout());
    }
    safe_free(buf);		/* Allocated by gopherSendRequest. */
    return 0;
}

/* This will be called when connect completes. Write request. */
int gopherSendRequest(fd, data)
int fd;
GopherData *data;
{
#define CR '\015'
#define LF '\012'
    int len;
    char query[MAX_URL];
    char *buf = (char *) xcalloc(1, MAX_URL);

    if (data->type_id == GOPHER_CSO) {
	sscanf(data->request, "?%s", query);
	len = strlen(query) + 15;
	sprintf(buf, "query %s%c%cquit%c%c", query, CR, LF, CR, LF);
    } else if (data->type_id == GOPHER_INDEX) {
	char *c_ptr = strchr(data->request, '?');
	if (c_ptr) {
	    *c_ptr = '\t';
	}
	len = strlen(data->request) + 3;
	sprintf(buf, "%s%c%c", data->request, CR, LF);
    } else {
	len = strlen(data->request) + 3;
	sprintf(buf, "%s%c%c", data->request, CR, LF);
    }

    debug(5, "gopherSendRequest - fd: %d\n", fd);
    icpWrite(fd, buf, len, 30, gopherSendComplete, data);
    return 0;
}

int gopherStart(unusedfd, url, entry)
int unusedfd;
char *url;
StoreEntry *entry;
{
    /* Create state structure. */
    int sock, status;
    GopherData *data = (GopherData *) xmalloc(sizeof(GopherData));

    memset(data, '\0', sizeof(GopherData));
    data->tail = data->buf;	/* XXX: huh? DRH */
    data->len = 0;
    data->data_in = 0;
    data->entry = entry;

    debug(3, "gopherStart - url: %s\n", url);

    /* Parse url. */
    if (gopher_url_parser(url, data->host, &data->port,
			  &data->type_id, data->request)) {
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		entry->url,
		entry->url,
		"Gopher",
		208,
		"Invalid URL syntax: Cannot parse.",
		"Contact your system adminstrator for further help.",
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(entry, tmp_error_buf);
	safe_free(data);
	return COMM_ERROR;
    }
    /* Create socket. */
    sock = comm_open(COMM_NONBLOCKING, 0, 0);

    /* check if IP is already in cache. It must be. 
       It should be done before this route is called. 
       Otherwise, we cannot check return code for connect. */
    if (!ipcache_gethostbyname(data->host)) {
	debug(4, "gopherStart: Called without IP entry in ipcache. OR lookup failed.\n");
	comm_close(sock);
	sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		entry->url,
		entry->url,
		"Gopher",
		202,
		"DNS name lookup failure",
		dns_error_message,
		HARVEST_VERSION,
		comm_hostname());
	storeAbort(entry, tmp_error_buf);
	safe_free(data);
	return COMM_ERROR;
    }
    if (((data->type_id == GOPHER_INDEX) || (data->type_id == GOPHER_CSO))
	&& (strchr(data->request, '?') == NULL)
	&& (BIT_TEST(data->entry->flag, REQ_HTML))) {
	/* Index URL without query word */
	/* We have to generate search page back to client. No need for connection */
	gopherMimeCreate(data);

	if (data->type_id == GOPHER_INDEX) {
	    data->conversion = HTML_INDEX_PAGE;
	} else {
	    if (data->type_id == GOPHER_CSO) {
		data->conversion = HTML_CSO_PAGE;
	    } else {
		data->conversion = HTML_INDEX_PAGE;
	    }
	}
	gopherToHTML(data, NULL, 0);
	storeComplete(data->entry);
	safe_free(data);
	return COMM_OK;
    }
    /* Open connection. */
    if ((status = comm_connect(sock, data->host, data->port)) != 0) {
	if (status != EINPROGRESS) {
	    comm_close(sock);
	    sprintf(tmp_error_buf, CACHED_RETRIEVE_ERROR_MSG,
		    entry->url,
		    entry->url,
		    "Gopher",
		    204,
		    "Cannot connect to the original site",
		    "The remote site may be down.",
		    HARVEST_VERSION,
		    comm_hostname());
	    storeAbort(entry, tmp_error_buf);
	    safe_free(data);
	    return COMM_ERROR;
	} else {
	    debug(5, "startGopher - conn %d EINPROGRESS\n", sock);
	}
    }
    /* Install connection complete handler. */
    comm_set_select_handler(sock, COMM_SELECT_LIFETIME,
			    (PF) gopherLifetimeExpire, (caddr_t) data);
    comm_set_select_handler(sock, COMM_SELECT_WRITE,
			    (PF) gopherSendRequest, (caddr_t) data);

    return COMM_OK;
}
