static char rcsid[] = "$Id: icp.c,v 1.98 1995/08/21 21:19:43 duane Exp $";
/* 
 *  File:         icp.c
 *  Description:  Implementation of binary cache protocol.
 *  Author:       John Noll, USC
 *                Dean Povey, University of Queensland
 *  Created:      Wed Mar 30 16:25:01 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>		/* for time(); */
#include <limits.h>		/* for USHRT_MAX */
#include <sys/errno.h>		/* for EWOULDBLOCK */
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "ansihelp.h"		/* goes first */
#include "debug.h"
#include "comm.h"
#include "proto.h"		/* for  caddr_t */
#include "store.h"
#include "stat.h"
#include "icp.h"
#include "stack.h"
#include "mime.h"
#include "cache_cf.h"
#include "recursive.h"

#ifdef SUPPORT_ICPV1_4
#include "icpv1.4.h"
#include "proto_v1.4.h"
#endif

extern Stack free_pkt_list;
extern int errno;
extern unsigned long nconn;

#if !defined(__FreeBSD__)
extern char *sys_errlist[];
#endif
extern int *fd_lifetime;

static int asciiPortNum = 0;
static char *_host_url;
static struct in_addr tmp_in_addr;

icpUdpData *UdpQueue = NULL;
icpUdpData *AppendUdp();
#define ICP_MAX_UDP_SIZE 4096

typedef int (*complete_handler) _PARAMS((int fd, char *buf, int size, int errflag, caddr_t data));
typedef struct ireadd {
    int fd;
    int binary_mode;
    char *buf;
    long size;
    long offset;
    int timeout;		/* XXX Not used at present. */
    time_t time;
    complete_handler handler;
    caddr_t client_data;
} icpReadWriteData;

typedef struct iwd {
    icp_common_t header;    /* Allows access to previous header */
    u_num32 query_host;
    int binary_mode;        /* If FALSE, transfer in ascii mode */
    char *url;
    char *type;         /* GET, POST, ... */
    int type_id;
    char *mime_hdr;     /* Mime header */
    int html_request;
    StoreEntry *entry;
    long offset;
} icpStateData;

/* A couple useful macros */
#define asciiMode(X)  (!(X->binary_mode))
#define binaryMode(X) ((X->binary_mode))

/* Local functions */
void icpHandleStore _PARAMS((int fd, StoreEntry * entry, icpStateData * state));
void icpHandleStoreComplete _PARAMS((int fd, char *buf, int size, int errflag, icpStateData * state));
int icpGetOptions _PARAMS((caddr_t buf));

/* XXX Bits set in return value from icpGetOPtions for various options */
#define BIT_OPT_NOP 0x0
#define BIT_OPT_NOPROXY 0x1
#define BIT_OPT_RECURSE 0x2


/* Save the ascii port num, so we can tell what error msg to send */
void icpSaveAsciiPort(port)
int port;
{
    asciiPortNum = port;

#ifdef SUPPORT_ICPV1_4
	/* Also save the port in the ICPv1.4 module */
	icpv1_4SaveAsciiPort(port);
#endif 
}

/* Read from FD. */
int icpHandleRead(fd, rw_state_machine)
int fd;
icpReadWriteData *rw_state_machine;
{
    int len = read(fd, rw_state_machine->buf + rw_state_machine->offset,
		   rw_state_machine->size - rw_state_machine->offset);

    if (len <= 0) {
	if (errno != EWOULDBLOCK || len == 0) {
	    /* Len == 0 means connection closed; otherwise,  would not have been
	       called by comm_select(). */
	    debug(1, "icpHandleRead: read failure: FD %d: %s\n",
		fd, len == 0 ? "connection closed" : sys_errlist[errno]);
	    rw_state_machine->handler(fd, rw_state_machine->buf,
				    rw_state_machine->offset, COMM_ERROR,
				      rw_state_machine->client_data);
	    safe_free(rw_state_machine);
	    return COMM_ERROR;
	}
    }
    rw_state_machine->offset += len;

    if (asciiMode(rw_state_machine)) {
	/* Check for \r\n delimiting end of ascii transmission */
	if (rw_state_machine->buf[rw_state_machine->offset - 2] == '\r' &&
	    rw_state_machine->buf[rw_state_machine->offset - 1] == '\n') {
	    rw_state_machine->handler(fd, rw_state_machine->buf,
				      rw_state_machine->offset,
				 COMM_OK, rw_state_machine->client_data);

	    safe_free(rw_state_machine);
	} else
	    comm_set_select_handler(fd, COMM_SELECT_READ,
				    (PF) icpHandleRead,
				    (caddr_t) rw_state_machine);
    } else if (rw_state_machine->offset < rw_state_machine->size) {
	comm_set_select_handler(fd, COMM_SELECT_READ,
				(PF) icpHandleRead,
				(caddr_t) rw_state_machine);
    } else {
	rw_state_machine->handler(fd, rw_state_machine->buf,
				  rw_state_machine->offset,
				  COMM_OK, rw_state_machine->client_data);

	/* We've read.  Toss state machine */
	safe_free(rw_state_machine);
    }
    return COMM_OK;
}

/* Select for reading on FD, until SIZE bytes are received.  Call
   HANDLER when complete. */
void icpRead(fd, bin_mode, buf, size, timeout, handler, client_data)
int fd;
int bin_mode;
char *buf;
int size;
int timeout;
int (*handler) _PARAMS((int fd, char *buf, int size, int errflag, caddr_t data));
caddr_t client_data;
{
    icpReadWriteData *data = (icpReadWriteData *) xcalloc(1, sizeof(icpReadWriteData));
    data->fd = fd;
    data->buf = buf;
    data->size = size;
    data->offset = 0;
    data->handler = handler;
    data->timeout = timeout;
    data->time = time(0);
    data->client_data = client_data;
    data->binary_mode = bin_mode;
    comm_set_select_handler(fd, COMM_SELECT_READ, (PF) icpHandleRead, (caddr_t) data);
}

/* Write to FD. */
void icpHandleWrite(fd, rwsm)
int fd;
icpReadWriteData *rwsm;
{
    int len = 0;

    debug(5, "icpHandleWrite: FD %d: off %d: sz %d.\n",
	  fd, rwsm->offset, rwsm->size);

    len = write(fd, rwsm->buf + rwsm->offset, rwsm->size - rwsm->offset);
    if (len <= 0) {

	if (len == 0) {
	    /* check if it's because no data to write or connection got closed */
	    if (rwsm->size - rwsm->offset == 0)
		rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_OK, rwsm->client_data);
	    else {
		debug(2, "icpHandleWrite: FD %d: write failure: connection closed.\n", fd);
		rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_ERROR, rwsm->client_data);
	    }
	    safe_free(rwsm);
	} else {
	    /* error case */
	    if (errno != EWOULDBLOCK) {
		debug(2, "icpHandleWrite: FD %d: write failure: %s.\n",
		      fd, sys_errlist[errno]);
		rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_ERROR, rwsm->client_data);
		safe_free(rwsm);
	    }
	}

    } else {
	rwsm->offset += len;
	if (rwsm->offset < rwsm->size) {
	    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpHandleWrite, (caddr_t) rwsm);
	} else {
	    rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_OK, rwsm->client_data);
	    safe_free(rwsm);
	}
    }
}

/* Select for Writing on FD, until SIZE bytes are sent.  Call
   HANDLER when complete. */
void icpWrite(fd, buf, size, timeout, handler, client_data)
int fd;
char *buf;
int size;
int timeout;
int (*handler) _PARAMS((int fd, char *buf, int size, int errflag, caddr_t data));
caddr_t client_data;
{
    icpReadWriteData *data;

    debug(5, "icpWrite: FD %d: sz %d: tout %d: hndl %p: data %p.\n",
	  fd, size, timeout, handler, client_data);

    data = (icpReadWriteData *) xcalloc(1, sizeof(icpReadWriteData));
    data->fd = fd;
    data->buf = buf;
    data->size = size;
    data->offset = 0;
    data->handler = handler;
    data->timeout = timeout;
    data->time = time(0);
    data->client_data = client_data;
    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpHandleWrite, (caddr_t) data);
}

void icpSendERRORComplete(fd, buf, size, errflag, state)
int fd;
char *buf;
int size;
int errflag;
caddr_t state;
{
    icpStateData *icp_state = (icpStateData *) state;

    debug(4, "icpSendERRORComplete: FD %d: sz %d: err %d.\n",
	  fd, size, errflag);

	if (icp_state)
	{
    	if (icp_state->entry) {
		storeUnregister(icp_state->entry, fd);
		storeUnlockObject(icp_state->entry);
		}
	}

    (void) comm_close(fd);
    safe_free(buf);
    if (icp_state) {
	safe_free(icp_state->url);
	safe_free(icp_state->type);
	safe_free(icp_state->mime_hdr);
	safe_free(icp_state);
    }
}

/* Send ERROR message. */
int icpSendERROR(fd, errorCode, msg, state)
int fd;
int errorCode;
char *msg;
icpStateData *state;
{
    char *buf = NULL, *p = NULL;
    unsigned short tmp_error = 0;
    icp_common_t tmp_header;
    int buf_len = 0, port = 0;

    port = comm_port(fd);
    debug(4, "icpSendERROR: code %d: port %d: msg: %1.80\n",
	  errorCode, port, msg);

    if (port == COMM_ERROR) {
	/* This file descriptor isn't bound to a socket anymore */
	/* Probably, this has timed out or somthing and someone 
	   called comm_cleanup_fd_entry.  We're getting a bug where
	   the select_loop table is also wrong.  Let's debug by purify
	   and intentionally create a memory leak by simply deallocating
	   the fd by brute force */
	debug(1, "icpSendERROR: COMM_ERROR msg: %1.80s\n", msg);
	/* We need to free up some state and free the handler,  I assume  */
	icpSendERRORComplete(fd, NULL, NULL, COMM_ERROR, state);
	return COMM_ERROR;
    } else if (port == asciiPortNum) {	/* Error message for the ascii port */
	buf_len = strlen(msg) + 1;	/* XXX: buf_len includes \0? */
	buf = (char *) xmalloc(buf_len);
	memset(buf, '\0', buf_len);
	strcpy(buf, msg);
    } else {			/* Default to binary error message */
	buf_len = strlen(msg) + 1;
	buf_len += sizeof(icp_common_t);
	buf_len += sizeof(unsigned short);
	buf = (char *) xmalloc(buf_len);
	memset(buf, '\0', buf_len);

	memset((char *) &tmp_header, '\0', sizeof(icp_common_t));
	tmp_header.opcode = ICP_OP_ERR;
	tmp_header.length = buf_len;
	tmp_header.offset = sizeof(icp_common_t);

	p = buf;
	memcpy(p, (char *) &tmp_header, sizeof(icp_common_t));
	p += sizeof(icp_common_t);

	tmp_error = htons(errorCode);
	memcpy(p, (char *) &tmp_error, sizeof(unsigned short));
	p += sizeof(unsigned short);

	memcpy(p, msg, strlen(msg) + 1);	/* include \0 */
    }
    icpWrite(fd, buf, buf_len, 30, icpSendERRORComplete, state);
    return COMM_OK;
}

/* Send available data from an object in the cache.  This is called either
   on select for  write or directly by icpHandleStore. */

int icpSendMoreData(fd, state)
int fd;
icpStateData *state;
{
    char *buf, *p;
    StoreEntry *entry = state->entry;
    int result = COMM_ERROR, max_len = 0;
    icp_common_t tmp_header, *header = &state->header;
    int buf_len, retval_ignored, len;

    debug(5, "icpSendMoreData: <URL:%s> sz %d: len %d: off %d.\n",
     entry->url, entry->object_len, entry->e_current_len, state->offset);

    if (!empty_stack(&free_pkt_list)) {
	pop(&free_pkt_list, &buf);
	p = buf;
    } else {
	p = buf = (char *) xcalloc(1, USHRT_MAX);
    }

    /* Set maxlen to largest amount of data w/o header
       place p pointing to beginning of data portion of message */

    if (asciiMode(state))	/* No header for ascii mode */
	buf_len = 0;
    else if ((state->offset == 0) && (header->opcode != ICP_OP_DATABEG))
	/* This is a DATA BEGIN message */
	buf_len = ICP_HDR_SZ + 2 * sizeof(u_num32);
    else
	buf_len = ICP_HDR_SZ;

    max_len = USHRT_MAX - buf_len;
    /* Should limit max_len to something like 1.5x last successful write */
    p += buf_len;

    retval_ignored = storeCopy(state->entry, state->offset, max_len, p, &len);

    buf_len += len;

    if ((state->offset == 0) && (header->opcode != ICP_OP_DATABEG)) {
	header->opcode = ICP_OP_DATABEG;
    } else if ((entry->e_current_len == entry->object_len) &&
	       ((entry->object_len - state->offset) == len)) {
	/* No more data; this is the last message. */
	header->opcode = ICP_OP_DATAEND;
    } else {
	/* We know there is more data to come. */
	header->opcode = ICP_OP_DATA;
    }
    debug(6, "icpSendMoreData: opcode %d: len %d.\n",
	  header->opcode, entry->object_len);

    header->length = buf_len;

    if (binaryMode(state)) {	/* Copy in a header */
	p = buf;

	tmp_header.opcode = header->opcode;
	tmp_header.version = header->version;
	tmp_header.length = htons(header->length);
	tmp_header.reqnum = htonl(header->reqnum);
	memcpy(tmp_header.auth, header->auth, sizeof(header->auth));
	memset((char *) &(tmp_header.padding), 0, 3);
	tmp_header.shostid = htonl(header->shostid);
	tmp_header.offset = htons(sizeof(icp_common_t));
	tmp_header.nr_options = 0;

	memcpy(p, (char *) &tmp_header, sizeof(icp_common_t));
	p += sizeof(icp_common_t);
	if (header->opcode == ICP_OP_DATABEG) {
	    icp_datab_t msg;
	    msg.db_ttl = htonl(entry->ttl);
	    msg.db_ts = htonl(entry->timestamp);
	    msg.db_size = htonl(entry->object_len);
	    memcpy(p, (char *) &msg, 3 * sizeof(u_num32));
	} else {
	    u_num32 tmp = htonl(state->offset);
	    memcpy(p, (char *) &tmp, sizeof(u_num32));
	}
    }
    state->offset += len;

    /* Do this here, so HandleStoreComplete can tell whether more data 
       needs to be sent. */
    icpWrite(fd, buf, buf_len, 30, icpHandleStoreComplete, state);
    result = COMM_OK;
    return result;
}

/* Called by storage manager when more data arrives from source. 
   Starts send sequence for a new batch of data. */
void icpHandleStore(fd, entry, state)
int fd;
StoreEntry *entry;
icpStateData *state;
{
    debug(5, "icpHandleStore: FD %d: off %d: <URL:%s>\n",
	  fd, state->offset, entry->url);

    if (entry->status == STORE_ABORTED) {
	icpSendERROR(fd, ICP_ERROR_TIMEDOUT, entry->e_abort_msg, state);
    } else {
	state->entry = entry;
	icpSendMoreData(fd, state);
    }
}

void icpHandleStoreComplete(fd, buf, size, errflag, state)
int fd;
char *buf;
int size;
int errflag;
icpStateData *state;
{
    icp_common_t *header = &state->header;

    debug(5, "icpHandleStoreComplete: FD %d: sz %d: err %d: off %d: len %d: tsmp %d: lref %d.\n",
	  fd, size, errflag,
	  state->offset, state->entry->object_len,
	  state->entry->timestamp, state->entry->lastref);


/*  free(buf); */
    push(&free_pkt_list, buf);

    if (errflag) {
	/* if runs in quick abort mode, set flag to tell 
	   fetching module to abort the fetching */
	if (getQuickAbort()) {
	    BIT_SET(state->entry->flag, CLIENT_ABORT_REQUEST);
	    tmp_in_addr.s_addr = (long) 0;	/* XXX Need a fix here */
	    CacheInfo->log_append(CacheInfo,
				  state->url,
				  inet_ntoa(tmp_in_addr),
				  state->entry->e_current_len,
				  "CLIENT_ABORT",
				  state->type);
	}
	storeUnregister(state->entry, fd);
	storeUnlockObject(state->entry);
	comm_close(fd);

	safe_free(state->url);
	safe_free(state->type);
	safe_free(state->mime_hdr);
	safe_free(state);
    } else if (state->offset < state->entry->e_current_len) {
	/* More data available locally; write it now */
	icpSendMoreData(fd, state);
    } else
	/* We're finished case */ if ((binaryMode(state) && (header->opcode == ICP_OP_DATAEND)) ||
     (asciiMode(state) && (state->offset == state->entry->object_len))) {
	storeUnregister(state->entry, fd);
	storeUnlockObject(state->entry);
	comm_close(fd);
	safe_free(state->url);
	safe_free(state->type);
	safe_free(state->mime_hdr);
	safe_free(state);
    } else
	/* We're in binary mode and we owe a DATAEND */ if (binaryMode(state) && (header->opcode != ICP_OP_DATAEND) &&
			   (state->offset == state->entry->object_len)) {
	icpSendMoreData(fd, state);
    } else {
	/* More data will be coming from primary server; register with 
	   storage manager. */
	storeRegister(state->entry, fd, icpHandleStore, state);
    }
}

int icpDoQuery(fd, data)
int fd;
icpStateData *data;
{
    /* XXX not implemented over tcp. */
    icpSendERROR(fd, ICP_ERROR_INTERNAL, "not implemented over tcp", data);
    return COMM_OK;
}

void icpProcessSEND(fd, usm)
int fd;
icpStateData *usm;
{
    char *url = usm->url;
    char *key;
    char *mime_hdr = usm->mime_hdr;
    StoreEntry *entry;

#define MAIL_SENDS 100
    static int sends = 0;
    extern int do_mail;

    debug(4, "icpProcessSEND: %s <URL:%s>\n", usm->type, url);

    if (usm->type_id == REQUEST_OP_GET) {
	key = url;
    } else {
	key = storeGenerateKey(usm->url, usm->type_id);
    }

    entry = storeGet(key);

    /* Yuk.  Seems like we should optimize this away... */
    if (getMailTrace() && (0 == do_mail) && (++sends > MAIL_SENDS)) {
	debug(2, "icpProcessSEND: scheduling trace mail to UC-Boulder.\n");
	do_mail = 1;
    }
    if (entry) {
	if (storeEntryValidToSend(entry) &&
	    (mime_hdr && !mime_refresh_request(mime_hdr))) {
	    debug(4, "icpProcessSEND: sending from store.\n");

	    /* STAT */
	    tmp_in_addr.s_addr = htonl(usm->header.shostid);
	    CacheInfo->log_append(CacheInfo, entry->url, inet_ntoa(tmp_in_addr),
				entry->object_len, "TCP_HIT", usm->type);
	    CacheInfo->proto_hit(CacheInfo, CacheInfo->proto_id(entry->url));

	    /* Reset header for reply. */
	    memset(&usm->header, 0, sizeof(icp_common_t));
	    usm->header.version = ICP_VERSION_CURRENT;
	    usm->header.reqnum = 0;
	    usm->header.shostid = 0;
		usm->header.offset = sizeof(icp_common_t);
	    usm->entry = entry;
	    usm->offset = 0;

	    /* Send reply. */
	    storeLockObject(entry);
	    icpSendMoreData(fd, usm);
	} else {
	    /* STAT */
	    tmp_in_addr.s_addr = htonl(usm->header.shostid);
	    CacheInfo->log_append(CacheInfo, url, inet_ntoa(tmp_in_addr), 0, "TCP_MISS_TTL", usm->type);
	    CacheInfo->proto_miss(CacheInfo, CacheInfo->proto_id(url));

	    icpProcessSENDA(fd, usm);
	}
    } else {
	/* STAT */
	tmp_in_addr.s_addr = htonl(usm->header.shostid);
	CacheInfo->log_append(CacheInfo, url, inet_ntoa(tmp_in_addr), 0, "TCP_MISS", usm->type);
	CacheInfo->proto_miss(CacheInfo, CacheInfo->proto_id(url));

	icpProcessSENDA(fd, usm);
    }
}

int icpProcessSENDA(fd, usm)
int fd;
icpStateData *usm;
{
    char *url = usm->url;
    char *key;
    char *type = usm->type;
    char *mime_hdr = usm->mime_hdr;
    StoreEntry *entry;

    debug(4, "icpProcessSENDA: %s <URL:%s>\n", type, url);
    debug(10, "icpProcessSENDA: mime_hdr: '%s'\n", mime_hdr);

    if (usm->type_id == REQUEST_OP_GET) {
	key = url;
    } else {
	key = storeGenerateKey(usm->url, usm->type_id);
    }

    entry = storeGet(key);

    if (entry) {
	/* get rid of the old entry */
	if (storeEntryLocked(entry)) {
	    /* change original hash key to get out of the new object's way */
	    if (storeOriginalKey(entry)) {
		/* protect infinite number of key changing */
		storeChangeKey(entry);
		debug(4, "icpProcessSENDA: Key Change: '%s' <URL:%s>\n",
		      entry->key, entry->url);
	    } else {
		debug(2, "icpProcessSENDA: Object located by changed key?\n");
	    }
	    BIT_SET(entry->flag, RELEASE_REQUEST);


	} else {
	    storeRelease(entry);
	}
	entry = storeAdd(url, type, mime_hdr, proto_cachable(url, type, mime_hdr), usm->html_request, usm->type_id);
    } else {
	entry = storeAdd(url, type, mime_hdr, proto_cachable(url, type, mime_hdr), usm->html_request, usm->type_id);
    }

    storeLockObject(entry);

    /*Reset header fields for  reply. */
    memset(&usm->header, 0, sizeof(icp_common_t));
    usm->header.version = ICP_VERSION_CURRENT;
    usm->header.reqnum = 0;
    usm->header.shostid = 0;
	usm->header.offset = htons(sizeof(icp_common_t));
    usm->entry = entry;
    usm->offset = 0;

    /* Register with storage manager to receive updates when data comes in. */
    storeRegister(entry, fd, icpHandleStore, usm);

    return (protoDispatch(fd, url, usm->entry));
}

void icpProcessUrl(fd, buf, size, flag, usm)
int fd;
char *buf;
int size;
int flag;
icpStateData *usm;
{
    if (flag || size < usm->header.length - sizeof(icp_common_t)) {
	debug(1, "icpProcessUrl: failure trying to read host id.\n");
	safe_free(buf);
	icpSendERROR(fd, ICP_ERROR_INTERNAL, "error reading host id", usm);
    } else {
	/* Ext/ract hostid. */
	memcpy(&usm->query_host, buf, sizeof(u_num32));
	usm->url = (char *) strdup(buf + sizeof(u_num32));
	usm->type = strdup("GET");
	usm->mime_hdr = NULL;

	safe_free(buf);

	/* Process request. */
	if (usm->header.opcode == ICP_OP_SEND) {
	   debug(5, "icpProcessUrl: processing ICP_OP_SEND\n");
	    icpProcessSEND(fd, usm);
	} else if (usm->header.opcode == ICP_OP_SENDA) {
	    debug(5, "icpProcessUrl: processing ICP_OP_SENDA\n");
	    icpProcessSENDA(fd, usm);
	} else if (usm->header.opcode == ICP_OP_QUERY) {
	    debug(5, "icpProcessUrl: processing ICP_OP_QUERY\n");
	    icpDoQuery(fd, usm);
	} else {
	    debug(1, "icpProcessUrl: Invalid OPCODE: %d.\n", usm->header.opcode);
	}
    }
}

int icpProcessHeader(fd, buf, size, flag, data)
int fd;
char *buf;
int size;
int flag;
icpStateData *data;
{
    int result = COMM_ERROR;

    debug(4, "icpProcessHeader: FD %d.\n", fd);

    if (flag || size < sizeof(icp_common_t)) {
	debug(1, "icpProcessHeader: FD %d: header read failure.\n", fd);
	icpSendERROR(fd, ICP_ERROR_INTERNAL, "error reading header", data);
	/* safe_free(buf);      Don't free this, it belongs to icpStateData */
	result = COMM_ERROR;
    } else {
	short op = ntohs(data->header.opcode);
	if (op == ICP_OP_SEND || op == ICP_OP_SENDA || op == ICP_OP_QUERY) {
	    /* Read query host id & url. */
	    int buf_size;
	    char *buf;

	    icp_common_t *hp = &data->header;
	    hp->opcode = op;
	    /* XXX Do these macros work ok in this fashion? */
	    hp->version = hp->version;
	    hp->length = ntohs(hp->length);
	    hp->reqnum = ntohl(hp->reqnum);
	    hp->shostid = ntohl(hp->shostid);
		hp->nr_options = ntohl(hp->nr_options);

		/* Process Options */

	    /* Allocate buffer for  hostid and url. */
	    buf_size = hp->length - hp->offset;
	    buf = (char *) xcalloc(buf_size, sizeof(char));

	    /* Schedule read of host id and url. */
	    (void) icpRead(fd, TRUE, buf, buf_size, 30, icpProcessUrl, data);
	} else {
	    debug(1, "icpProcessHeader: FD %d: invalid OPCODE: %d\n", fd, op);
	    icpSendERROR(fd, ICP_ERROR_INTERNAL, "invalid opcode", data);
	    result = COMM_ERROR;
	}
    }
    return result;
}

int icpHandleTcp(sock, notused)
int sock;
caddr_t notused;
{
    int fd;
    icpStateData *data;
    if ((fd = comm_accept(sock)) < 0) {
	debug(1, "icphandleTcp: accept failure: %d\n", sock);
	/* XXX Should we close and exit? */
	return -1;
    }
    debug(4, "icpHandleTcp: FD %d: accept succeeded.\n", fd);
    /* Schedule read of message header. */
    data = (icpStateData *) xcalloc(1, sizeof(icpStateData));
    data->binary_mode = 1;
    (void) icpRead(fd, TRUE, &data->header, sizeof(icp_common_t), 30, icpProcessHeader, data);

    /* Reschedule accept handler. */
    comm_set_select_handler(sock, COMM_SELECT_READ, icpHandleTcp, 0);
    return 0;
}


int icpUdpReply(fd, queue)
int fd;
icpUdpData *queue;
{
    int result = COMM_OK;
    icpUdpData *dp;

    dp = UdpQueue;
    /* Disable handler, in case of errors. */
    comm_set_select_handler(fd, COMM_SELECT_WRITE, 0, 0);

    if (comm_udp_sendto(fd, &queue->address, sizeof(struct sockaddr_in),
			queue->msg, queue->len) < 0) {
	debug(1, "icpUdpReply: error sending\n");
	result = COMM_ERROR;
    }
    /* Don't close socket, as we need it for new incoming requests.  Just remove
       handler for writes. */
    UdpQueue = UdpQueue->next;
    if (!UdpQueue)
	comm_set_select_handler(fd, COMM_SELECT_WRITE, 0, 0);
    else
	comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpUdpReply,
				(caddr_t) UdpQueue);
    safe_free(dp->msg);
    safe_free(dp);

    return result;
}

/*
 * icpUdpMiss - Return a miss message
 */
int icpUdpMiss(fd, url, old_buf, from)
int fd;
char *url;
caddr_t old_buf;
struct sockaddr_in *from;
{
    char *buf = NULL;
    int buf_len = sizeof(icp_common_t) + strlen(url) + 1;
    icp_common_t *headerp = (icp_common_t *) old_buf;
    icpUdpData *data = (icpUdpData *) xmalloc(sizeof(icpUdpData));
    struct sockaddr_in our_socket_name;
    int sock_name_length = sizeof(our_socket_name);

    if (getsockname(fd, (struct sockaddr *) &our_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "icpUdpMiss: FD %d: getsockname failure: %s\n",
	      fd, strerror(errno));
    }
    debug(5, "icpUdpMiss: FD %d: %s: <URL:%s>\n", fd,
	  inet_ntoa(our_socket_name.sin_addr), url);

	/* Copy the address to send the packet to */
    memcpy(&data->address, from, sizeof(struct sockaddr_in));

	/* Allocate memory for the packet */
    buf = (char *) xcalloc(1, buf_len);

	/* Copy query header (should be identical except opcode and packet length) 
	 */ 
	memcpy(buf, old_buf, headerp->offset);

	/* Change opcode */ 
	headerp = (icp_common_t *) buf;
	headerp->opcode = ICP_OP_MISS;

	/* Set packet length */
	headerp->length = htons(headerp->offset + strlen(url) +1);

	/* Copy the url */
    memcpy(buf + headerp->offset, url, strlen(url) + 1);
    data->msg = buf;
    data->len = buf_len;
    UdpQueue = AppendUdp(data, UdpQueue);

    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpUdpReply,
			    (caddr_t) UdpQueue);

    return COMM_OK;
}

/* 
 * icpUdpHit - Return a hit message
 * 
 */
int icpUdpHit(fd, url, old_buf, entry, from)
int fd;
char *url;
caddr_t old_buf;
StorePtrEntry *entry;
struct sockaddr_in *from;
{
    char *buf, *tmp_buf;
    icp_common_t *headerp = (icp_common_t *) old_buf;
    icpUdpData *data = (icpUdpData *) xcalloc(1, sizeof(icpUdpData));
	StorePointer *tmp=0;
    unsigned short url_len=0;
	unsigned short nr_ptrs=0;
	unsigned short buf_len=0;
	struct sockaddr_in our_socket_name;
    int sock_name_length = sizeof(our_socket_name);

	/* Log the call */
    if (getsockname(fd, (struct sockaddr *) &our_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "icpUdpHit: FD %d: getsockname failure: %s\n",
	      fd, strerror(errno));
    }
    debug(5, "icpUdpHit: FD %d: %s: <URL:%s>\n", fd,
	  inet_ntoa(our_socket_name.sin_addr), url);

	
	/* Copy the address to send the packet */
    memcpy(&data->address, from, sizeof(struct sockaddr_in));

	/* 
	 * Check to see if we have a saved copy of the hit data, if not
	 * work out how big this packet needs to be
	 */
	if (!entry->hit_data)
	{
		/* Get source url_len */
		url_len = htons(strlen(url) + 1);

		entry->hit_data_len =  sizeof(u_num32) /* Nr of ptrs + src url len */
        		+ ntohs(url_len);              /* Source url */
   ;

		/* Work out the size of the ptr data */
		tmp = entry->ptr;
		while(tmp)
		{
			entry->hit_data_len += 
				+ sizeof(u_num32)           /* TTL */
				+ sizeof(unsigned short)    /* Pointer length */
				+ strlen(tmp->ptr_url) + 1; /* Pointer url */
	
			tmp = tmp->next;
		}

	}

	/* Allocate memory for the packet */
	buf_len = headerp->offset  /* header */
		+ entry->hit_data_len; /* Hit data */
	buf = xcalloc(buf_len, 1);


	/* 
	 * Copy the header (should be identical to query packet) 
	 * XXX Although we may want to add some options later
	 */
	memcpy(buf, old_buf, headerp->offset); 

	/* Change opcode */ 
	headerp = (icp_common_t *) buf;	
	headerp->opcode = ICP_OP_HIT;

	/* Set length */
	headerp->length = htons(buf_len);

	/* Check to see if we have saved hit data, if not create some */
	if (!entry->hit_data)
	{
		if (!entry->local)
		{
			/* Save this data for next time */
			entry->hit_data = xcalloc(entry->hit_data_len, 1);
		}
		else
		{
			/* XXX The hit data is thrown away */
			entry->hit_data = buf + headerp->offset;
		}

		/* Copy the number of pointers */
		tmp_buf = entry->hit_data;
		nr_ptrs = htons(entry->nr_ptrs);
		memcpy(tmp_buf, &nr_ptrs, sizeof(unsigned short));
	
		/* Copy the length of the URL */
		tmp_buf += sizeof(unsigned short);
		memcpy(tmp_buf, &url_len, sizeof(unsigned short));
		
		/* Copy the URL */
    	tmp_buf += sizeof(unsigned short); 
    	memcpy(tmp_buf, url, ntohs(url_len));
		tmp_buf += ntohs(url_len);
	
		/* Copy the hit list */
		tmp = entry->ptr;
		while(tmp)
		{
			/* Copy the ttl */
			memcpy(tmp_buf, &tmp->ttl, sizeof(u_num32));
			tmp_buf += sizeof(u_num32);
	
        	/* Get the size of the pointer URL */
        	url_len = htons(strlen(tmp->ptr_url) + 1);
	
			/* Copy the size of the pointer URL */
			memcpy(tmp_buf, &url_len, sizeof(unsigned short));
			tmp_buf += sizeof(unsigned short);
	
			/* Copy the pointer URL */
			memcpy(tmp_buf, tmp->ptr_url, ntohs(url_len));
			tmp_buf += ntohs(url_len);
	
			tmp = tmp->next;
		}
	
	}

	if (!entry->local)
		/* Copy the hit data */
		memcpy(buf + headerp->offset, entry->hit_data, entry->hit_data_len);

	/* Append data to the UDP queue */ 
    data->msg = buf;
    data->len = buf_len;
    UdpQueue = AppendUdp(data, UdpQueue);

    debug(4, "icpUdpHit: FD %d: from %s: sz %d: <URL:%s>\n", fd,
	  inet_ntoa(from->sin_addr), buf_len, url);

    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpUdpReply,
			    (caddr_t) UdpQueue);

    return COMM_OK;
}

int icpUdpRecursiveReply(fd, url, buf, size)
int fd;
char *url;
caddr_t buf;
int size;
{
	icpUdpData *data=0; 
	recQueryData *query=0;
	recQueryData *tmp=0;
	icp_common_t *headerp = (icp_common_t *)buf;

	/* Get the registered hosts for this query */
	query = recGetRequest(url);
	if (!query)
	{
		/* XXX - Log an error (possibly this is due to a delayed packet) 
		 * causing a timeout and freeing up the query before it had a chance
		 * to complete
		 */
		return 0;
	}
	else 
	{
		/* 
		 * Cycle through the recursive query data and send a reply to each
		 * registered host  
		 */
		do
		{
			/* Allocate memory for the Udp data */
			data = (icpUdpData *) xcalloc(1, sizeof(icpUdpData));

  			/* Append data to the UDP queue */
			memcpy(&data->address, &query->reply_to, 
				sizeof(struct sockaddr_in));

			/* Copy the buffer */
    		data->msg = xcalloc(1, size);
			memcpy(data->msg, buf, size);
    		data->len = size;

			/* 
			 * Change the request id and host in the header (note these
			 * should already be in network byte order 
			 */
			headerp = (icp_common_t *) data->msg;
			headerp->shostid = query->host; 
			headerp->reqnum = query->reqnum;

    		UdpQueue = AppendUdp(data, UdpQueue);

			/* XXX Log this ? */
    		comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpUdpReply,
                (caddr_t) UdpQueue);

			query=query->next;	
		} while (query);
	}

	/* Remove the Recursive Query data */
	recReleaseRequest(url);

	return COMM_OK;
	
}  

/* 
 * If the recursive resolution times out then we send a miss back to the 
 * requester.  
 * XXX - This should be negatively cached 
 */
int icpUdpRecursiveTimeout(fd, url)
int fd;
char *url;
{
    int buf_len = sizeof(icp_common_t) + sizeof(unsigned char) + 
		strlen(url) + 1;
	caddr_t buf;
    icp_common_t *headerp = 0; 
	unsigned char recurse_option = htons(ICP_OPT_RECURSE);

    debug(5, "icpUdpRecursiveTimeout: FD %d: <URL:%s>\n", fd, url);

	/* Allocate memory for the packet */
    buf = (char *) xcalloc(1, buf_len);
	headerp = (icp_common_t *) buf;

	/* Create Packet to send back */
	headerp->version = ICP_VERSION_CURRENT;
	headerp->opcode = ICP_OP_MISS;
	headerp->length = htons(buf_len);
	memset(headerp->padding, 0, sizeof(headerp->padding));
	headerp->offset = sizeof(icp_common_t) + sizeof(unsigned short); 
	headerp->nr_options = 1; /* XXX May add more later */
	memcpy(buf + sizeof(icp_common_t), &recurse_option, sizeof(unsigned char));

	/* Copy the url */
    memcpy(buf + headerp->offset, url, strlen(url) + 1);

	/* Send the Miss packet */
	return icpUdpRecursiveReply(fd, url, buf, buf_len);
}

int icpUdpSend(fd, url, reqheaderp, from, opcode, options)
int fd;
char *url;
icp_common_t *reqheaderp;
struct sockaddr_in *from;
icp_opcode opcode;
unsigned char options;
{
    char *buf;
    int buf_len = sizeof(icp_common_t) + strlen(url) + 1;
    icp_common_t *headerp;
    icpUdpData *data = (icpUdpData *) xmalloc(sizeof(icpUdpData));
    struct sockaddr_in our_socket_name;
    int sock_name_length = sizeof(our_socket_name);
	unsigned char recurse_option=ICP_OPT_RECURSE;
	char *urloffset;

    if (getsockname(fd, (struct sockaddr *) &our_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "icpUdpSend: FD %d: getsockname failure: %s\n",
	      fd, strerror(errno));
    }
    memset(data, '\0', sizeof(icpUdpData));
    memcpy(&data->address, from, sizeof(struct sockaddr_in));

    if (opcode == ICP_OP_QUERY)
	{
		buf_len += sizeof(u_num32);

		/* XXX currently only RECURSE works for queries */
		if (options)
			buf_len += sizeof(unsigned char);
	}
    buf = (char *) xcalloc(1, buf_len);

    headerp = (icp_common_t *) buf;
    headerp->opcode = opcode;
    headerp->version = ICP_VERSION_CURRENT;
    headerp->length = htons(buf_len);

	/* 
	 * XXX Currently this will fuck up on little endian machines 
	 * Fix neighborsUdpPing
	 */
    headerp->reqnum = reqheaderp->reqnum;  
    headerp->shostid = reqheaderp->shostid;

/*  memcpy(headerp->auth, , ICP_AUTH_SIZE); */
	if (!options)
	{
		headerp->offset = sizeof(icp_common_t);
		headerp->nr_options = 0;
	}
	else
	{
		/* XXX currently only RECURSE is valid so this is ok */
		headerp->offset = sizeof(icp_common_t) + sizeof(unsigned char);
		headerp->nr_options = 1;
		memcpy(buf + sizeof(icp_common_t), &recurse_option, 
			sizeof(unsigned char));
	}
	
    urloffset = buf + headerp->offset; 

    if (opcode == ICP_OP_QUERY)
	urloffset += sizeof(u_num32);
    /* it's already zero filled by xcalloc */
    memcpy(urloffset, url, strlen(url));
    data->msg = buf;
    data->len = buf_len;

    UdpQueue = AppendUdp(data, UdpQueue);

    debug(4, "icpUdpSend: op %d: from %s: sz %d: <URL:%s>\n", opcode,
	  inet_ntoa(from->sin_addr), buf_len, url);

    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpUdpReply,
			    (caddr_t) UdpQueue);

    return COMM_OK;
}

int icpUdpForward(fd, buf, buf_len, from)
int fd;
caddr_t buf;
int buf_len;
struct sockaddr_in *from;
{
    icpUdpData *data = (icpUdpData *) xmalloc(sizeof(icpUdpData));
    struct sockaddr_in our_socket_name;
    int sock_name_length = sizeof(our_socket_name);

    if (getsockname(fd, (struct sockaddr *) &our_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "icpUdpSend: FD %d: getsockname failure: %s\n",
	      fd, strerror(errno));
    }
    memset(data, '\0', sizeof(icpUdpData));
    memcpy(&data->address, from, sizeof(struct sockaddr_in));

    data->msg = buf;
    data->len = buf_len;

    UdpQueue = AppendUdp(data, UdpQueue);

    debug(4, "icpUdpForward: from %s: sz %d:\n", 
		inet_ntoa(from->sin_addr), buf_len);

    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpUdpReply,
			    (caddr_t) UdpQueue);

    return COMM_OK;
}


int icpHandleUdp(sock, not_used)
int sock;
caddr_t not_used;
{

    int result = 0;
    struct sockaddr_in from;
    int from_len;
    char buf[ICP_MAX_UDP_SIZE];
    int len;
	unsigned short url_len=0;
    icp_common_t header, *headerp;
    StorePtrEntry *ptr_entry;
	StoreEntry *entry;
    char *url;

    from_len = sizeof(from);
    memset(&from, 0, from_len);
    /* zero filled to make sure url is terminated. */
    memset(&buf, 0, ICP_MAX_UDP_SIZE);


    if ((len = comm_udp_recv(sock, buf, ICP_MAX_UDP_SIZE - 1, &from, 
		&from_len)) < 0) {
	debug(1, "icpHandleUdp: FD %d: error receiving.\n", sock);
	/* don't exit here. soft error. */
	/* exit(1); */
	comm_set_select_handler(sock, COMM_SELECT_READ, icpHandleUdp, 0);
	return result;
    }
    debug(4, "icpHandleUdp: FD %d: received %d bytes from %s.\n", sock, len,
	  inet_ntoa(from.sin_addr));

    if (len < sizeof(icp_common_t) 
#ifdef SUPPORT_ICPV1_4
		&& len < sizeof(icpv1_4_common_t)
#endif
	) {
	debug(4, "icpHandleUdp: Bad UDP packet, Ignored. Size: %d < Header: %d\n",
	      len, sizeof(icp_common_t));
    } else {

	/* Get fields from incoming message. */
	headerp = (icp_common_t *) buf;
	header.opcode = headerp->opcode;
	header.version = ntohs(headerp->version);;

#ifdef SUPPORT_ICPV1_4
	if (header.version < ICP_VERSION_CURRENT)
	{
		/* 
		 * The older version of the cached code didn't seem to check version
         * numbers, so we'll assume that 0,1 and 2 can all be thrown at the
		 * old version of this code
         */

		/* Process ICP v1.4 message */
		result =  icpv1_4HandleUdp(sock, buf, &from, len);
    	comm_set_select_handler(sock, COMM_SELECT_READ, icpHandleUdp, 0);
    	return result;
	}
#endif

	header.length = ntohs(headerp->length);
	header.reqnum = ntohs(headerp->reqnum);
	header.shostid = ntohs(headerp->shostid);
	/*  memcpy(headerp->auth, , ICP_AUTH_SIZE); */
	/*  Ignore padding for the time being */ 
	header.offset = ntohs(headerp->offset);
	header.nr_options = ntohs(headerp->nr_options);

	switch (header.opcode) {
	case ICP_OP_QUERY:
	    {
		if (len < sizeof(icp_common_t)) {
		    /* at least it has to have \0 as a URL */
		    debug(4, "icpHandleUdp: Bad ICP_OP_QUERY packet, Ignored.\n");
		    debug(4, "Size: %d < Min: %d\n", len,
			  sizeof(icp_common_t) + sizeof(u_num32) + 1);
		    break;
		}
		url = buf + header.offset + sizeof(u_num32);

		if (ip_access_check(from.sin_addr, ip_access_control_list) == IP_DENY) {
		    debug(2, "icpHandleUdp: Access Denied for %s.\n", inet_ntoa(from.sin_addr));
		    CacheInfo->log_append(CacheInfo, "ACCESS_DENIED", inet_ntoa(from.sin_addr), 0, "UDP_DENIED", "GET");
		} else {
		    ptr_entry = storeCheckCache(url);
		    debug(5, "icpHandleUdp: OPCODE: ICP_OP_QUERY\n");

		    if (ptr_entry) {
			/* Send "HIT" message. */

			/* XXX - Need to check if this entry is pending or not, if so
			 * then some way for the reply to be sent back to the originator
			 * needs to be resolved 
			 */

			/* STAT - NB: Need to append stats for all hits */
			CacheInfo->log_append(CacheInfo, ptr_entry->url, 
						inet_ntoa(from.sin_addr), 
						ptr_entry->object_len, 
						"UDP_HIT", 
						"GET");
					CacheInfo->proto_hit(CacheInfo, 
						CacheInfo->proto_id(ptr_entry->url));

					/* Send the Hit message */
					icpUdpHit(sock, url, buf, ptr_entry, &from);
		    } else {

				if (icpGetOptions(buf) & BIT_OPT_RECURSE)
				{
					/* Register recursive query */
					recRegisterRequest(url, headerp->shostid, 
						headerp->reqnum, &from);

					/* Do a recursive query */
					protoDispatchRecursiveQuery(sock, url, buf);
				}
				else
				{
					/* Send "MISS" message. */
					/* STAT */
					CacheInfo->log_append(CacheInfo, url, 
						inet_ntoa(from.sin_addr),
					      0, "UDP_MISS", "GET");
					CacheInfo->proto_miss(CacheInfo, CacheInfo->proto_id(url));

					icpUdpMiss(sock, url, buf, &from);
		    	}
			}
	    }
		}
	    break;

	case ICP_OP_HIT:
		/* Get the url */
		url = buf + header.offset + (sizeof(u_num32));

		/* Check the options - if this is the result of a recursive query, then
		 * get the associate pointer
		 * XXX DO some logging
		 */
		if (icpGetOptions(buf) & BIT_OPT_RECURSE)
		{
			icpUdpRecursiveReply(sock, url, buf, len);
		}
		else
		{
			/* Get the store entry for this URL */
	    	entry = storeGet(url);
	    	debug(4, "icpHandleUdp: HIT for %s on <URL:%s>.\n", 
				inet_ntoa(from.sin_addr), url);
	    	if (!entry) {
			debug(4, 
		"icpHandleUdp: received ICP_OP_HIT but StoreEntry is null. Ignored.\n");
	    	} else {
			neighborsUdpAck(sock, url, buf, &from, entry);
	    	}
		}
	    break;

	case ICP_OP_SECHO:
		/* NB: This should never be received for a recursive query */
	    url = buf + header.offset;		/* XXX url known to be 0 term'd ? */
	    entry = storeGet(url);
	    debug(4, "icpHandleUdp: SECHO for %s on <URL:%s>\n", inet_ntoa(from.sin_addr), url);
	    if (!entry) {
		debug(4, "icpHandleUdp: received ICP_OP_SECHO but StoreEntry is null. Ignored.\n");
	    } else {
		neighborsUdpAck(sock, url, &header, &from, entry);
	    }
	    break;

	case ICP_OP_MISS:
	    url = buf + header.offset;		/* XXX url known to be 0 term'd ? */

	    /* Check the options - if this is the result of a recursive query, then
		 * get the associate pointer
		 * XXX DO some logging
		 */
		if (icpGetOptions(buf) & BIT_OPT_RECURSE)
		{
			icpUdpRecursiveReply(sock, url, buf, len);
		}
		else
		{
			entry = storeGet(url);
	    	debug(4, "icpHandleUdp: MISS for %s on <URL:%s>\n", 
				inet_ntoa(from.sin_addr), url);
	    	if (!entry) {
			debug(4, 
	"icpHandleUdp: received ICP_OP_MISS but StoreEntry is null. Ignored.\n");
	    	} else {
			neighborsUdpAck(sock, url, &header, &from, entry);
	    	}
		}
	    break;

	case ICP_OP_ADD:
		icpAddPointers(buf, len);
		break;
	case ICP_OP_DELETE:
		icpDeletePointers(buf, len);
		break;
	default:
	    debug(4, "icpHandleUdp: OPCODE: %d\n", header.opcode);
	    break;
	}
    }
    comm_set_select_handler(sock, COMM_SELECT_READ, icpHandleUdp, 0);
    return result;
}

void asciiStripNewLine(s)
char *s;
{
    if ((s[strlen(s) - 1] == '\n') || (s[strlen(s) - 1] == '\r'))
	s[strlen(s) - 1] = '\0';
    if ((s[strlen(s) - 1] == '\n') || (s[strlen(s) - 1] == '\r'))
	s[strlen(s) - 1] = '\0';
}

#ifdef APPEND_DOMAIN
static char *
 do_append_domain(url)
char *url;
{
    char *b, *e, *p, *u;
    int lo, ln;

    if (!(b = strstr(url, "://")))
	return (url);
    b += 3;
    if (!(e = strchr(b, '/')))
	e = b + strlen(b);
    if ((p = strchr(b, '@')) && p < e)
	b = p + 1;
    if ((p = strchr(b, ':')) && p < e)
	e = p;
    if ((p = strchr(b, '.')) && p < e)
	return (url);
    lo = strlen(url);
    ln = lo + append_domain_len;
    if (!(u = xrealloc(url, ln + 1)))
	return (url);
    b = u + lo;
    p = u + ln;
    e = u + (e - url);
    while (b >= e)
	*p-- = *b--;
    strncpy(e, append_domain, append_domain_len);
    return (u);
}
#endif				/* APPEND_DOMAIN */


char *
 parseAsciiUrl(input, type, mime_hdr, html_request, type_id)
char *input;
char **type;
char **mime_hdr;
int *html_request;
int *type_id;
{
    char *cmd, *url = NULL;
    char *ind, *token, *ptr;

    cmd = (char *) strtok(input, " ");
    if (cmd && isascii(cmd[0])) {

	if ((strncmp(cmd, "get", 3) == 0) || (strncmp(cmd, "GET", 3) == 0)) {
	    if (strncmp(cmd, "GET", 3) == 0)
		*html_request = 1;
	    else
		*html_request = 0;

	    cmd = (char *) strtok(NULL, "\n\r ");
	    *type = strdup("GET");
	    *type_id = REQUEST_OP_GET;
	    token = (char *) strtok(NULL, "");
	    if (token)
		*mime_hdr = strdup(token);
	    else
		*mime_hdr = strdup(" ");

	} else if ((strncmp(cmd, "post", 4) == 0) ||
		   (strncmp(cmd, "POST", 4) == 0)) {
	    *html_request = 1;
	    cmd = (char *) strtok(NULL, "\n\r ");
	    *type = strdup((char *) "POST");
	    *type_id = REQUEST_OP_POST;
	    token = (char *) strtok(NULL, "");
	    if (token)
		*mime_hdr = strdup(token);
	    else
		*mime_hdr = strdup(" ");

	} else if ((strncmp(cmd, "head", 4) == 0) ||
		   (strncmp(cmd, "HEAD", 4) == 0)) {
	    *html_request = 1;
	    cmd = (char *) strtok(NULL, "\n\r ");
	    *type = strdup((char *) "HEAD");
	    *type_id = REQUEST_OP_HEAD;
	    token = (char *) strtok(NULL, "");
	    if (token)
		*mime_hdr = strdup(token);
	    else
		*mime_hdr = strdup(" ");

	} else {
	    *type = strdup("GET");
	    *type_id = REQUEST_OP_GET;
	    *html_request = 0;
	    *mime_hdr = strdup(" ");
	}

#ifdef APPEND_DOMAIN
	if (append_domain && url)
	    url = do_append_domain(url);
#endif				/* APPEND_DOMAIN */


	if (cmd) {		/* A hack to avoid death from null request */
	    asciiStripNewLine(cmd);

	    if (ptr = strchr(cmd, '#'))		/* Lousy hack to remove   */
		*ptr = '\0';	/* HTML bookmarks */

	    /* a hack to get around If-Modified-Since: Header */
	    if (mime_hdr)
		mime_process(*mime_hdr);

	    /* see if we running in httpd_accel_mode, if so got to convert it to URL */

	    if (httpd_accel_mode) {
		if (cmd[0] == '/') {
		    /* got to convert it */
		    url = (char *) xcalloc(1, strlen(cmd) + strlen(httpd_accel_prefix) + 1);
		    sprintf(url, "%s%s", httpd_accel_prefix, cmd);
		} else {
		    url = (char *) strdup(cmd);
		}
	    } else {
		url = (char *) strdup(cmd);
	    }
	}
    }
    debug(5, "parseAsciiUrl: %s: <URL:%s>: mime_hdr '%s'.\n", *type, url ? url : "NULL", *mime_hdr);
    return url;
}

/* Read incoming request and perform requested operation. */
void asciiProcessInput(fd, buf, size, flag, astm)
int fd;
char *buf;
int size;
int flag;
icpStateData *astm;
{
    debug(4, "asciiProcessInput: FD %d: reading request...\n", fd);

    if (flag == COMM_OK) {
	if ((astm->url = parseAsciiUrl(buf, &(astm->type), &(astm->mime_hdr),
			     &(astm->html_request), &(astm->type_id)))) {
	    astm->offset = strlen(astm->url);
	    icpProcessSEND(fd, astm);
	    safe_free(buf);
	} else {
	    debug(1, "asciiProcessInput: FD %d: read failure.\n", fd);
	    icpSendERROR(fd, ICP_ERROR_INTERNAL, "error reading request", astm);
	    safe_free(buf);
	}
    } else {
	/* connection closed by foreign host */
	comm_close(fd);
	safe_free(buf);
	safe_free(astm);
    }
}

/* general lifetime handler for ascii connection */
int asciiConnLifetimeHandle(fd, data)
int fd;
caddr_t data;
{
    icpStateData *astm = (icpStateData *) data;
    PF handler;
    caddr_t client_data;
    icpReadWriteData *rw_state;

    debug(2, "asciiConnLifetimeHandle: Socket: %d lifetime is expired. Free up data structure.\n", fd);

    /* look up for write handler if it's there */
    handler = NULL;
    client_data = NULL;
    comm_get_select_handler(fd, COMM_SELECT_WRITE, (PF *) & handler, (caddr_t *) & client_data);
    if ((handler != NULL) && (client_data != NULL)) {
	rw_state = (icpReadWriteData *) client_data;
	if (rw_state->buf)
	    safe_free(rw_state->buf);
	safe_free(rw_state);
    }
    /* look up for read handler if it's there */
    handler = NULL;
    client_data = NULL;
    comm_get_select_handler(fd, COMM_SELECT_READ, (PF *) & handler, (caddr_t *) & client_data);
    if ((handler != NULL) && (client_data != NULL)) {
	rw_state = (icpReadWriteData *) client_data;
	if (rw_state->buf)
	    safe_free(rw_state->buf);
	safe_free(rw_state);
    }
    if (astm->url) {
	/* try to unregister us from dnsserver pending list */
	/* this call will take care of that */
	protoUndispatch(fd, astm->url);
    }
    if (astm->entry) {
	storeUnregister(astm->entry, fd);
	storeUnlockObject(astm->entry);
    }
    safe_free(astm->url);
    safe_free(astm->mime_hdr);
    safe_free(astm->type);
    safe_free(astm);
    comm_close(fd);
}

/* Handle a new connection on ascii input socket. */
int asciiHandleConn(sock, notused)
int sock;
caddr_t notused;
{
    int fd;
    char tmpbuf[512];
    icpStateData *astm;
    struct sockaddr_in peer_socket_name;
    int sock_name_length = sizeof(peer_socket_name);

    if ((fd = comm_accept(sock)) < 0) {
	debug(1, "asciiHandleConn: FD %d: accept failure: %s\n",
	      sock, sys_errlist[errno]);
	/* XXX Should we close and exit? */
	return -1;
    }
    /* set the hardwired lifetime */
    fd_lifetime[fd] = time(NULL) + getClientLifetime();
    nconn++;

    debug(4, "asciiHandleConn: FD %d: accepted (lifetime %d).\n", fd,
	  fd_lifetime[fd]);

    if (getpeername(fd, (struct sockaddr *) &peer_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "asciiHandleConn: FD %d: getpeername failure: %s.\n",
	      fd, strerror(errno));
    }
    if (ip_access_check(peer_socket_name.sin_addr, ip_access_control_list) == IP_DENY) {
	debug(2, "asciiHandleConn: %s: Access denied.\n", inet_ntoa(peer_socket_name.sin_addr));
	CacheInfo->log_append(CacheInfo, "ACCESS_DENIED", inet_ntoa(peer_socket_name.sin_addr), 0, "TCP_DENIED", "GET");
	sprintf(tmpbuf, "ACCESS DENIED\n\nYour IP address (%s) is not authorized to access cached at %s.\n\n",
		inet_ntoa(peer_socket_name.sin_addr), comm_hostname());
	write(fd, tmpbuf, strlen(tmpbuf));
	comm_close(fd);
    } else {
	astm = (icpStateData *) xcalloc(1, sizeof(icpStateData));
	astm->binary_mode = FALSE;
	astm->url = (char *) xcalloc(1, 4096);
	astm->header.shostid = htonl(peer_socket_name.sin_addr.s_addr);
	comm_set_select_handler(fd, COMM_SELECT_LIFETIME, (PF) asciiConnLifetimeHandle,
				(caddr_t) astm);
	icpRead(fd, FALSE, astm->url, 4065, 30, asciiProcessInput, astm);
    }

    comm_set_select_handler(sock, COMM_SELECT_READ, asciiHandleConn, 0);
    return 0;

}


icpUdpData *
 AppendUdp(data, q)
icpUdpData *data;
icpUdpData *q;
{
    icpUdpData *p;

    data->next = NULL;
    if (q == NULL) {
	q = data;
	data->next = NULL;
    } else {
	p = q;
	while (p->next) {
	    p = p->next;
	}
	p->next = data;
    }
    return (q);
}

int icpGetOptions(buf)
caddr_t buf;
{

	/* 
	 * XXX Process the options and return an integer with bits set indicating
	 * appropriate option. This only supports options with no data (and only 
	 * 16 options [depending on sizeof(int)])
	 */
	int opts=0;
	icp_common_t *header = (icp_common_t *)buf;
	int options[ICP_OPT_END]={BIT_OPT_NOP, BIT_OPT_NOPROXY, BIT_OPT_RECURSE};
	caddr_t tmp = buf + sizeof(icp_common_t);
	int i;

	for (i=0; i<header->nr_options; i++)
	{
		unsigned char option_nr = *((unsigned char *)tmp);
		
		/* Make sure the option is supported */
		if (option_nr > ICP_OPT_HIGHEST)
		{
			return -1;
		}
		else
		{
			/* Set the bit for this option */ 
			opts |= options[option_nr]; 
		}

		/* Point to the next option */
		tmp += sizeof(unsigned char);
	}

	return opts;
}


void icpSetHostUrl(hostname, ascii_port, udp_port, tcp_port)
	char *hostname;
	int ascii_port;
	int udp_port;
	int tcp_port;
{
	_host_url = xcalloc(strlen(hostname) + 40, 1); /* XXX should be enough */ 
	sprintf(_host_url, "%s:%d:%d:%d", hostname, ascii_port, udp_port, tcp_port);
	return;
}

char *icpGetHostUrl()
{
	return _host_url;
}

int icpAddPointers(buf, buf_len)
	caddr_t buf;
	int buf_len;
{
	/* Pointers to bits of packet data */
	icp_common_t *headerp = (icp_common_t *) buf;
	char *tmp =  buf + headerp->offset; 

	/* Fields in the packet */
	unsigned char scope;
	unsigned char nr_items;
	u_num32 timestamp;
	u_num32 ttl;
	u_num32 size;
	unsigned short url_len;
	char *src_url;
	char *ptr_url;

	/* XXX Some logging should go on in here */

	/* Get header data */
	memcpy(&scope, tmp, sizeof(unsigned char));
	tmp += sizeof(unsigned char);
	memcpy(&nr_items, tmp, sizeof(unsigned char));
	tmp += sizeof(unsigned char);

	/* Add the pointers */
	while(nr_items--)
	{
		/* Get header data */
		memcpy(&size, tmp, sizeof(u_num32)); 
		tmp += sizeof(u_num32);
		memcpy(&ttl, tmp, sizeof(u_num32)); 
		tmp += sizeof(u_num32);
		memcpy(&timestamp, tmp, sizeof(u_num32)); 
		tmp += sizeof(u_num32);

		/* Get src url */
		memcpy(&url_len, tmp, sizeof(unsigned short));
		tmp = src_url = tmp + sizeof(unsigned short);
		tmp += ntohs(url_len);

		/* Get pointer url */
		memcpy(&url_len, tmp, sizeof(unsigned short));
		tmp = ptr_url = tmp + sizeof(unsigned short);
		tmp += ntohs(url_len);

		storeAddPointer(src_url, ptr_url, ntohl(ttl), ntohl(timestamp), 
			ntohl(size));
	}
	
	/* Decrement scope */
	scope--;
	if (scope)
	{
		/* Forward the packet */
		memcpy(buf + headerp->offset, &scope, sizeof(unsigned char)); 
		return neighborsUdpAdvertise(buf, buf_len);
	}
	else
	{
		return COMM_OK;
	}
}

int icpDeletePointers(buf, buf_len)
	caddr_t buf;
	int buf_len;
{
	/* Pointers to bits of packet data */
	icp_common_t *headerp = (icp_common_t *) buf;
	char *tmp =  buf + headerp->offset; 

	/* Fields in the packet */
	unsigned char scope;
	unsigned char nr_items;
	u_num32 timestamp;
	u_num32 ttl;
	u_num32 size;
	unsigned short url_len;
	char *src_url;
	char *ptr_url;

	/* XXX Some logging should go on in here */

	/* Get header data */
	memcpy(&scope, tmp, sizeof(unsigned char));
	tmp += sizeof(unsigned char);
	memcpy(&nr_items, tmp, sizeof(unsigned char));
	tmp += sizeof(unsigned char);

	/* Delete the pointers */
	while(nr_items--)
	{
		/* Get src url */
		memcpy(&url_len, tmp, sizeof(unsigned short));
		tmp = src_url = tmp + sizeof(unsigned short);
		tmp += ntohs(url_len);

		/* Get pointer url */
		memcpy(&url_len, tmp, sizeof(unsigned short));
		tmp = ptr_url = tmp + sizeof(unsigned short);
		tmp += ntohs(url_len);

		storeDelPointer(src_url, ptr_url);
	}
	
	/* Decrement scope */
	scope--;
	if (scope)
	{
		/* Forward the packet */
		memcpy(buf + headerp->offset, &scope, sizeof(unsigned char)); 
		return neighborsUdpAdvertise(buf, buf_len);
	}
	else
	{
		return COMM_OK;
	}
}
