static char rcsid[] = "$Id: icpv1.4.c,v 1.98 1995/08/21 21:19:43 duane Exp $";
/* 
 *  File:         icpv1.4.c
 *  Description:  Implementation of binary cache protocol.
 *  Author:       John Noll, USC
 *  Created:      Wed Mar 30 16:25:01 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */

#ifdef SUPPORT_ICPV1_4

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>		/* for time(); */
#include <limits.h>		/* for USHRT_MAX */
#include <sys/errno.h>		/* for EWOULDBLOCK */
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "ansihelp.h"		/* goes first */
#include "debug.h"
#include "comm.h"
#include "proto.h"
#include "proto_v1.4.h"		/* for  caddr_t */
#include "store.h"
#include "stat.h"
#include "icpv1.4.h"
#include "stack.h"
#include "mime.h"
#include "cache_cf.h"

extern Stack free_pkt_list;
extern int errno;
extern unsigned long nconn;
#if !defined(__FreeBSD__)
extern char *sys_errlist[];
#endif
extern int *fd_lifetime;

static int asciiPortNum = 0;
static struct in_addr tmp_in_addr;

typedef struct iwd {
    icpv1_4_common_t header;	/* Allows access to previous header */
    u_num32 query_host;
    int binary_mode;		/* If FALSE, transfer in ascii mode */
    char *url;
    char *type;			/* GET, POST, ... */
    int type_id;
    char *mime_hdr;		/* Mime header */
    int html_request;
    StoreEntry *entry;
    long offset;
} icpv1_4StateData;


static icpv1_4UdpData *UdpQueue = NULL;
static icpv1_4UdpData *AppendUdp();
#define ICPV1_4_MAX_UDP_SIZE 4096

typedef int (*complete_handler) _PARAMS((int fd, char *buf, int size, int errflag, caddr_t data));
typedef struct ireadd {
    int fd;
    int binary_mode;
    char *buf;
    long size;
    long offset;
    int timeout;		/* XXX Not used at present. */
    time_t time;
    complete_handler handler;
    caddr_t client_data;
} icpv1_4ReadWriteData;

/* A couple useful macros */
#define asciiMode(X)  (!(X->binary_mode))
#define binaryMode(X) ((X->binary_mode))

/* Local functions */
void icpv1_4HandleStore _PARAMS((int fd, StoreEntry * entry, icpv1_4StateData * state));
void icpv1_4HandleStoreComplete _PARAMS((int fd, char *buf, int size, int errflag, icpv1_4StateData * state));

static icpv1_4UdpData *AppendUdp();

/* Save the ascii port num, so we can tell what error msg to send */
void icpv1_4SaveAsciiPort(port)
int port;
{
    asciiPortNum = port;
}

/* Read from FD. */
int icpv1_4HandleRead(fd, rw_state_machine)
int fd;
icpv1_4ReadWriteData *rw_state_machine;
{
    int len = read(fd, rw_state_machine->buf + rw_state_machine->offset,
		   rw_state_machine->size - rw_state_machine->offset);

    if (len <= 0) {
	if (errno != EWOULDBLOCK || len == 0) {
	    /* Len == 0 means connection closed; otherwise,  would not have been
	       called by comm_select(). */
	    debug(1, "icpv1_4HandleRead: read failure: FD %d: %s\n",
		fd, len == 0 ? "connection closed" : sys_errlist[errno]);
	    rw_state_machine->handler(fd, rw_state_machine->buf,
				    rw_state_machine->offset, COMM_ERROR,
				      rw_state_machine->client_data);
	    safe_free(rw_state_machine);
	    return COMM_ERROR;
	}
    }
    rw_state_machine->offset += len;

    if (asciiMode(rw_state_machine)) {
	/* Check for \r\n delimiting end of ascii transmission */
	if (rw_state_machine->buf[rw_state_machine->offset - 2] == '\r' &&
	    rw_state_machine->buf[rw_state_machine->offset - 1] == '\n') {
	    rw_state_machine->handler(fd, rw_state_machine->buf,
				      rw_state_machine->offset,
				 COMM_OK, rw_state_machine->client_data);

	    safe_free(rw_state_machine);
	} else
	    comm_set_select_handler(fd, COMM_SELECT_READ,
				    (PF) icpv1_4HandleRead,
				    (caddr_t) rw_state_machine);
    } else if (rw_state_machine->offset < rw_state_machine->size) {
	comm_set_select_handler(fd, COMM_SELECT_READ,
				(PF) icpv1_4HandleRead,
				(caddr_t) rw_state_machine);
    } else {
	rw_state_machine->handler(fd, rw_state_machine->buf,
				  rw_state_machine->offset,
				  COMM_OK, rw_state_machine->client_data);

	/* We've read.  Toss state machine */
	safe_free(rw_state_machine);
    }
    return COMM_OK;
}

/* Select for reading on FD, until SIZE bytes are received.  Call
   HANDLER when complete. */
void icpv1_4Read(fd, bin_mode, buf, size, timeout, handler, client_data)
int fd;
int bin_mode;
char *buf;
int size;
int timeout;
int (*handler) _PARAMS((int fd, char *buf, int size, int errflag, caddr_t data));
caddr_t client_data;
{
    icpv1_4ReadWriteData *data = (icpv1_4ReadWriteData *) xcalloc(1, sizeof(icpv1_4ReadWriteData));
    data->fd = fd;
    data->buf = buf;
    data->size = size;
    data->offset = 0;
    data->handler = handler;
    data->timeout = timeout;
    data->time = time(0);
    data->client_data = client_data;
    data->binary_mode = bin_mode;
    comm_set_select_handler(fd, COMM_SELECT_READ, (PF) icpv1_4HandleRead, (caddr_t) data);
}

/* Write to FD. */
void icpv1_4HandleWrite(fd, rwsm)
int fd;
icpv1_4ReadWriteData *rwsm;
{
    int len = 0;

    debug(5, "icpv1_4HandleWrite: FD %d: off %d: sz %d.\n",
	  fd, rwsm->offset, rwsm->size);

    len = write(fd, rwsm->buf + rwsm->offset, rwsm->size - rwsm->offset);
    if (len <= 0) {

	if (len == 0) {
	    /* check if it's because no data to write or connection got closed */
	    if (rwsm->size - rwsm->offset == 0)
		rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_OK, rwsm->client_data);
	    else {
		debug(2, "icpv1_4HandleWrite: FD %d: write failure: connection closed.\n", fd);
		rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_ERROR, rwsm->client_data);
	    }
	    safe_free(rwsm);
	} else {
	    /* error case */
	    if (errno != EWOULDBLOCK) {
		debug(2, "icpv1_4HandleWrite: FD %d: write failure: %s.\n",
		      fd, sys_errlist[errno]);
		rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_ERROR, rwsm->client_data);
		safe_free(rwsm);
	    }
	}

    } else {
	rwsm->offset += len;
	if (rwsm->offset < rwsm->size) {
	    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpv1_4HandleWrite, (caddr_t) rwsm);
	} else {
	    rwsm->handler(fd, rwsm->buf, rwsm->offset, COMM_OK, rwsm->client_data);
	    safe_free(rwsm);
	}
    }
}

/* Select for Writing on FD, until SIZE bytes are sent.  Call
   HANDLER when complete. */
void icpv1_4Write(fd, buf, size, timeout, handler, client_data)
int fd;
char *buf;
int size;
int timeout;
int (*handler) _PARAMS((int fd, char *buf, int size, int errflag, caddr_t data));
caddr_t client_data;
{
    icpv1_4ReadWriteData *data;

    debug(5, "icpv1_4Write: FD %d: sz %d: tout %d: hndl %p: data %p.\n",
	  fd, size, timeout, handler, client_data);

    data = (icpv1_4ReadWriteData *) xcalloc(1, sizeof(icpv1_4ReadWriteData));
    data->fd = fd;
    data->buf = buf;
    data->size = size;
    data->offset = 0;
    data->handler = handler;
    data->timeout = timeout;
    data->time = time(0);
    data->client_data = client_data;
    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpv1_4HandleWrite, (caddr_t) data);
}

void icpv1_4SendERRORComplete(fd, buf, size, errflag, state)
int fd;
char *buf;
int size;
int errflag;
caddr_t state;
{
    icpv1_4StateData *icpv1_4_state = (icpv1_4StateData *) state;

    debug(4, "icpv1_4SendERRORComplete: FD %d: sz %d: err %d.\n",
	  fd, size, errflag);
    if (icpv1_4_state->entry) {
	storeUnregister(icpv1_4_state->entry, fd);
	storeUnlockObject(icpv1_4_state->entry);
    }
    (void) comm_close(fd);
    safe_free(buf);
    if (icpv1_4_state) {
	safe_free(icpv1_4_state->url);
	safe_free(icpv1_4_state->type);
	safe_free(icpv1_4_state->mime_hdr);
	safe_free(icpv1_4_state);
    }
}

/* Send ERROR message. */
int icpv1_4SendERROR(fd, errorCode, msg, state)
int fd;
int errorCode;
char *msg;
icpv1_4StateData *state;
{
    char *buf = NULL, *p = NULL;
    unsigned short tmp_error = 0;
    icpv1_4_common_t tmp_header;
    int buf_len = 0, port = 0;

    port = comm_port(fd);
    debug(4, "icpv1_4SendERROR: code %d: port %d: msg: %1.80\n",
	  errorCode, port, msg);

    if (port == COMM_ERROR) {
	/* This file descriptor isn't bound to a socket anymore */
	/* Probably, this has timed out or somthing and someone 
	   called comm_cleanup_fd_entry.  We're getting a bug where
	   the select_loop table is also wrong.  Let's debug by purify
	   and intentionally create a memory leak by simply deallocating
	   the fd by brute force */
	debug(1, "icpv1_4SendERROR: COMM_ERROR msg: %1.80s\n", msg);
	/* We need to free up some state and free the handler,  I assume  */
	icpv1_4SendERRORComplete(fd, NULL, NULL, COMM_ERROR, state);
	return COMM_ERROR;
    } else if (port == asciiPortNum) {	/* Error message for the ascii port */
	buf_len = strlen(msg) + 1;	/* XXX: buf_len includes \0? */
	buf = (char *) xmalloc(buf_len);
	memset(buf, '\0', buf_len);
	strcpy(buf, msg);
    } else {			/* Default to binary error message */
	buf_len = strlen(msg) + 1;
	buf_len += sizeof(icpv1_4_common_t);
	buf_len += sizeof(unsigned short);
	buf = (char *) xmalloc(buf_len);
	memset(buf, '\0', buf_len);

	memset((char *) &tmp_header, '\0', sizeof(icpv1_4_common_t));
	tmp_header.opcode = ICPV1_4_OP_ERR;
	tmp_header.length = buf_len;

	p = buf;
	memcpy(p, (char *) &tmp_header, sizeof(icpv1_4_common_t));
	p += sizeof(icpv1_4_common_t);

	tmp_error = htons(errorCode);
	memcpy(p, (char *) &tmp_error, sizeof(unsigned short));
	p += sizeof(unsigned short);

	memcpy(p, msg, strlen(msg) + 1);	/* include \0 */
    }
    icpv1_4Write(fd, buf, buf_len, 30, icpv1_4SendERRORComplete, state);
    return COMM_OK;
}

/* Send available data from an object in the cache.  This is called either
   on select for  write or directly by icpv1_4HandleStore. */

int icpv1_4SendMoreData(fd, state)
int fd;
icpv1_4StateData *state;
{
    char *buf, *p;
    StoreEntry *entry = state->entry;
    int result = COMM_ERROR, max_len = 0;
    icpv1_4_common_t tmp_header, *header = &state->header;
    int buf_len, retval_ignored, len;

    debug(5, "icpv1_4SendMoreData: <URL:%s> sz %d: len %d: off %d.\n",
     entry->url, entry->object_len, entry->e_current_len, state->offset);

    if (!empty_stack(&free_pkt_list)) {
	pop(&free_pkt_list, &buf);
	p = buf;
    } else {
	p = buf = (char *) xcalloc(1, USHRT_MAX);
    }

    /* Set maxlen to largest amount of data w/o header
       place p pointing to beginning of data portion of message */

    if (asciiMode(state))	/* No header for ascii mode */
	buf_len = 0;
    else if ((state->offset == 0) && (header->opcode != ICPV1_4_OP_DATABEG))
	/* This is a DATA BEGIN message */
	buf_len = ICPV1_4_HDR_SZ + 2 * sizeof(u_num32);
    else
	buf_len = ICPV1_4_HDR_SZ;

    max_len = USHRT_MAX - buf_len;
    /* Should limit max_len to something like 1.5x last successful write */
    p += buf_len;

    retval_ignored = storeCopy(state->entry, state->offset, max_len, p, &len);

    buf_len += len;

    if ((state->offset == 0) && (header->opcode != ICPV1_4_OP_DATABEG)) {
	header->opcode = ICPV1_4_OP_DATABEG;
    } else if ((entry->e_current_len == entry->object_len) &&
	       ((entry->object_len - state->offset) == len)) {
	/* No more data; this is the last message. */
	header->opcode = ICPV1_4_OP_DATAEND;
    } else {
	/* We know there is more data to come. */
	header->opcode = ICPV1_4_OP_DATA;
    }
    debug(6, "icpv1_4SendMoreData: opcode %d: len %d.\n",
	  header->opcode, entry->object_len);

    header->length = buf_len;

    if (binaryMode(state)) {	/* Copy in a header */
	p = buf;

	tmp_header.opcode = header->opcode;
	tmp_header.version = header->version;
	tmp_header.length = htons(header->length);
	tmp_header.reqnum = htonl(header->reqnum);
	memcpy(tmp_header.auth, header->auth, sizeof(header->auth));
	tmp_header.shostid = htonl(header->shostid);

	memcpy(p, (char *) &tmp_header, sizeof(icpv1_4_common_t));
	p += sizeof(icpv1_4_common_t);
	if (header->opcode == ICPV1_4_OP_DATABEG) {
	    icpv1_4_datab_t msg;
	    msg.db_ttl = htonl(entry->ttl);
	    msg.db_ts = htonl(entry->timestamp);
	    msg.db_size = htonl(entry->object_len);
	    memcpy(p, (char *) &msg, 3 * sizeof(u_num32));
	} else {
	    u_num32 tmp = htonl(state->offset);
	    memcpy(p, (char *) &tmp, sizeof(u_num32));
	}
    }
    state->offset += len;

    /* Do this here, so HandleStoreComplete can tell whether more data 
       needs to be sent. */
    icpv1_4Write(fd, buf, buf_len, 30, icpv1_4HandleStoreComplete, state);
    result = COMM_OK;
    return result;
}

/* Called by storage manager when more data arrives from source. 
   Starts send sequence for a new batch of data. */
void icpv1_4HandleStore(fd, entry, state)
int fd;
StoreEntry *entry;
icpv1_4StateData *state;
{
    debug(5, "icpv1_4HandleStore: FD %d: off %d: <URL:%s>\n",
	  fd, state->offset, entry->url);

    if (entry->status == STORE_ABORTED) {
	icpv1_4SendERROR(fd, ICPV1_4_ERROR_TIMEDOUT, entry->e_abort_msg, state);
    } else {
	state->entry = entry;
	icpv1_4SendMoreData(fd, state);
    }
}

void icpv1_4HandleStoreComplete(fd, buf, size, errflag, state)
int fd;
char *buf;
int size;
int errflag;
icpv1_4StateData *state;
{
    icpv1_4_common_t *header = &state->header;

    debug(5, "icpv1_4HandleStoreComplete: FD %d: sz %d: err %d: off %d: len %d: tsmp %d: lref %d.\n",
	  fd, size, errflag,
	  state->offset, state->entry->object_len,
	  state->entry->timestamp, state->entry->lastref);


/*  free(buf); */
    push(&free_pkt_list, buf);

    if (errflag) {
	/* if runs in quick abort mode, set flag to tell 
	   fetching module to abort the fetching */
	if (getQuickAbort()) {
	    BIT_SET(state->entry->flag, CLIENT_ABORT_REQUEST);
	    tmp_in_addr.s_addr = (long) 0;	/* XXX Need a fix here */
	    CacheInfo->log_append(CacheInfo,
				  state->url,
				  inet_ntoa(tmp_in_addr),
				  state->entry->e_current_len,
				  "CLIENT_ABORT",
				  state->type);
	}
	storeUnregister(state->entry, fd);
	storeUnlockObject(state->entry);
	comm_close(fd);

	safe_free(state->url);
	safe_free(state->type);
	safe_free(state->mime_hdr);
	safe_free(state);
    } else if (state->offset < state->entry->e_current_len) {
	/* More data available locally; write it now */
	icpv1_4SendMoreData(fd, state);
    } else
	/* We're finished case */ if ((binaryMode(state) && (header->opcode == ICPV1_4_OP_DATAEND)) ||
     (asciiMode(state) && (state->offset == state->entry->object_len))) {
	storeUnregister(state->entry, fd);
	storeUnlockObject(state->entry);
	comm_close(fd);
	safe_free(state->url);
	safe_free(state->type);
	safe_free(state->mime_hdr);
	safe_free(state);
    } else
	/* We're in binary mode and we owe a DATAEND */ if (binaryMode(state) && (header->opcode != ICPV1_4_OP_DATAEND) &&
			   (state->offset == state->entry->object_len)) {
	icpv1_4SendMoreData(fd, state);
    } else {
	/* More data will be coming from primary server; register with 
	   storage manager. */
	storeRegister(state->entry, fd, icpv1_4HandleStore, state);
    }
}

int icpv1_4DoQuery(fd, data)
int fd;
icpv1_4StateData *data;
{
    /* XXX not implemented over tcp. */
    icpv1_4SendERROR(fd, ICPV1_4_ERROR_INTERNAL, "not implemented over tcp", data);
    return COMM_OK;
}

void icpv1_4ProcessSEND(fd, usm)
int fd;
icpv1_4StateData *usm;
{
    char *url = usm->url;
    char *key;
    char *mime_hdr = usm->mime_hdr;
    StoreEntry *entry;

#define MAIL_SENDS 100
    static int sends = 0;
    extern int do_mail;

    debug(4, "icpv1_4ProcessSEND: %s <URL:%s>\n", usm->type, url);

    if (usm->type_id == REQUEST_OP_GET) {
	key = url;
    } else {
	key = storeGenerateKey(usm->url, usm->type_id);
    }

    entry = storeGet(key);

    /* Yuk.  Seems like we should optimize this away... */
    if (getMailTrace() && (0 == do_mail) && (++sends > MAIL_SENDS)) {
	debug(2, "icpv1_4ProcessSEND: scheduling trace mail to UC-Boulder.\n");
	do_mail = 1;
    }
    if (entry) {
	if (storeEntryValidToSend(entry) &&
	    (mime_hdr && !mime_refresh_request(mime_hdr))) {
	    debug(4, "icpv1_4ProcessSEND: sending from store.\n");

	    /* STAT */
	    tmp_in_addr.s_addr = htonl(usm->header.shostid);
	    CacheInfo->log_append(CacheInfo, entry->url, inet_ntoa(tmp_in_addr),
				entry->object_len, "TCP_HIT", usm->type);
	    CacheInfo->proto_hit(CacheInfo, CacheInfo->proto_id(entry->url));

	    /* Reset header for reply. */
	    memset(&usm->header, 0, sizeof(icpv1_4_common_t));
	    usm->header.version = ICPV1_4_VERSION_CURRENT;
	    usm->header.reqnum = 0;
	    usm->header.shostid = 0;
	    usm->entry = entry;
	    usm->offset = 0;

	    /* Send reply. */
	    storeLockObject(entry);
	    icpv1_4SendMoreData(fd, usm);
	} else {
	    /* STAT */
	    tmp_in_addr.s_addr = htonl(usm->header.shostid);
	    CacheInfo->log_append(CacheInfo, url, inet_ntoa(tmp_in_addr), 0, "TCP_MISS_TTL", usm->type);
	    CacheInfo->proto_miss(CacheInfo, CacheInfo->proto_id(url));

	    icpv1_4ProcessSENDA(fd, usm);
	}
    } else {
	/* STAT */
	tmp_in_addr.s_addr = htonl(usm->header.shostid);
	CacheInfo->log_append(CacheInfo, url, inet_ntoa(tmp_in_addr), 0, "TCP_MISS", usm->type);
	CacheInfo->proto_miss(CacheInfo, CacheInfo->proto_id(url));

	icpv1_4ProcessSENDA(fd, usm);
    }
}

int icpv1_4ProcessSENDA(fd, usm)
int fd;
icpv1_4StateData *usm;
{
    char *url = usm->url;
    char *key;
    char *type = usm->type;
    char *mime_hdr = usm->mime_hdr;
    StoreEntry *entry;

    debug(4, "icpv1_4ProcessSENDA: %s <URL:%s>\n", type, url);
    debug(10, "icpv1_4ProcessSENDA: mime_hdr: '%s'\n", mime_hdr);

    if (usm->type_id == REQUEST_OP_GET) {
	key = url;
    } else {
	key = storeGenerateKey(usm->url, usm->type_id);
    }

    entry = storeGet(key);

    if (entry) {
	/* get rid of the old entry */
	if (storeEntryLocked(entry)) {
	    /* change original hash key to get out of the new object's way */
	    if (storeOriginalKey(entry)) {
		/* protect infinite number of key changing */
		storeChangeKey(entry);
		debug(4, "icpv1_4ProcessSENDA: Key Change: '%s' <URL:%s>\n",
		      entry->key, entry->url);
	    } else {
		debug(2, "icpv1_4ProcessSENDA: Object located by changed key?\n");
	    }
	    BIT_SET(entry->flag, RELEASE_REQUEST);


	} else {
	    storeRelease(entry);
	}
	entry = storeAdd(url, type, mime_hdr, proto_cachable(url, type, mime_hdr), usm->html_request, usm->type_id);
    } else {
	entry = storeAdd(url, type, mime_hdr, proto_cachable(url, type, mime_hdr), usm->html_request, usm->type_id);
    }

    storeLockObject(entry);

    /*Reset header fields for  reply. */
    memset(&usm->header, 0, sizeof(icpv1_4_common_t));
    usm->header.version = ICPV1_4_VERSION_CURRENT;
    usm->header.reqnum = 0;
    usm->header.shostid = 0;
    usm->entry = entry;
    usm->offset = 0;

    /* Register with storage manager to receive updates when data comes in. */
    storeRegister(entry, fd, icpv1_4HandleStore, usm);

    return (protoDispatch(fd, url, usm->entry));
}

void icpv1_4ProcessUrl(fd, buf, size, flag, usm)
int fd;
char *buf;
int size;
int flag;
icpv1_4StateData *usm;
{
    if (flag || size < usm->header.length - sizeof(icpv1_4_common_t)) {
	debug(1, "icpv1_4ProcessUrl: failure trying to read host id.\n");
	safe_free(buf);
	icpv1_4SendERROR(fd, ICPV1_4_ERROR_INTERNAL, "error reading host id", usm);
    } else {
	/* Extract hostid. */
	memcpy(&usm->query_host, buf, sizeof(u_num32));
	usm->url = (char *) strdup(buf + sizeof(u_num32));
	usm->type = strdup("GET");
	usm->mime_hdr = NULL;

	safe_free(buf);

	/* Process request. */
	if (usm->header.opcode == ICPV1_4_OP_SEND) {
	    debug(5, "icpv1_4ProcessUrl: processing ICPV1_4_OP_SEND\n");
	    icpv1_4ProcessSEND(fd, usm);
	} else if (usm->header.opcode == ICPV1_4_OP_SENDA) {
	    debug(5, "icpv1_4ProcessUrl: processing ICPV1_4_OP_SENDA\n");
	    icpv1_4ProcessSENDA(fd, usm);
	} else if (usm->header.opcode == ICPV1_4_OP_QUERY) {
	    debug(5, "icpv1_4ProcessUrl: processing ICPV1_4_OP_QUERY\n");
	    icpv1_4DoQuery(fd, usm);
	} else {
	    debug(1, "icpv1_4ProcessUrl: Invalid OPCODE: %d.\n", usm->header.opcode);
	}
    }
}

int icpv1_4ProcessHeader(fd, buf, size, flag, data)
int fd;
char *buf;
int size;
int flag;
icpv1_4StateData *data;
{
    int result = COMM_ERROR;

    debug(4, "icpv1_4ProcessHeader: FD %d.\n", fd);

    if (flag || size < sizeof(icpv1_4_common_t)) {
	debug(1, "icpv1_4ProcessHeader: FD %d: header read failure.\n", fd);
	icpv1_4SendERROR(fd, ICPV1_4_ERROR_INTERNAL, "error reading header", data);
	/* safe_free(buf);      Don't free this, it belongs to icpv1_4StateData */
	result = COMM_ERROR;
    } else {
	short op = ntohs(data->header.opcode);
	if (op == ICPV1_4_OP_SEND || op == ICPV1_4_OP_SENDA || op == ICPV1_4_OP_QUERY) {
	    /* Read query host id & url. */
	    int buf_size;
	    char *buf;

	    icpv1_4_common_t *hp = &data->header;
	    hp->opcode = op;
	    /* XXX Do these macros work ok in this fashion? */
	    hp->version = hp->version;
	    hp->length = ntohs(hp->length);
	    hp->reqnum = ntohl(hp->reqnum);
	    hp->shostid = ntohl(hp->shostid);

	    /* Allocate buffer for  hostid and url. */
	    buf_size = hp->length - sizeof(icpv1_4_common_t);
	    buf = (char *) xcalloc(buf_size, sizeof(char));

	    /* Schedule read of host id and url. */
	    (void) icpv1_4Read(fd, TRUE, buf, buf_size, 30, icpv1_4ProcessUrl, data);
	} else {
	    debug(1, "icpv1_4ProcessHeader: FD %d: invalid OPCODE: %d\n", fd, op);
	    icpv1_4SendERROR(fd, ICPV1_4_ERROR_INTERNAL, "invalid opcode", data);
	    result = COMM_ERROR;
	}
    }
    return result;
}

int icpv1_4HandleTcp(sock, notused)
int sock;
caddr_t notused;
{
    int fd;
    icpv1_4StateData *data;
    if ((fd = comm_accept(sock)) < 0) {
	debug(1, "icpv1_4handleTcp: accept failure: %d\n", sock);
	/* XXX Should we close and exit? */
	return -1;
    }
    debug(4, "icpv1_4HandleTcp: FD %d: accept succeeded.\n", fd);
    /* Schedule read of message header. */
    data = (icpv1_4StateData *) xcalloc(1, sizeof(icpv1_4StateData));
    data->binary_mode = 1;
    (void) icpv1_4Read(fd, TRUE, &data->header, sizeof(icpv1_4_common_t), 30, icpv1_4ProcessHeader, data);

    /* Reschedule accept handler. */
    comm_set_select_handler(sock, COMM_SELECT_READ, icpv1_4HandleTcp, 0);
    return 0;
}


int icpv1_4UdpReply(fd, queue)
int fd;
icpv1_4UdpData *queue;
{
    int result = COMM_OK;
    icpv1_4UdpData *dp;

    dp = UdpQueue;
    /* Disable handler, in case of errors. */
    comm_set_select_handler(fd, COMM_SELECT_WRITE, 0, 0);

    if (comm_udp_sendto(fd, &queue->address, sizeof(struct sockaddr_in),
			queue->msg, queue->len) < 0) {
	debug(1, "icpv1_4UdpReply: error sending\n");
	result = COMM_ERROR;
    }
    /* Don't close socket, as we need it for new incoming requests.  Just remove
       handler for writes. */
    UdpQueue = UdpQueue->next;
    if (!UdpQueue)
	comm_set_select_handler(fd, COMM_SELECT_WRITE, 0, 0);
    else
	comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpv1_4UdpReply,
				(caddr_t) UdpQueue);
    safe_free(dp->msg);
    safe_free(dp);

    return result;
}

int icpv1_4UdpMiss(fd, url, reqheaderp, from)
int fd;
char *url;
icpv1_4_common_t *reqheaderp;
struct sockaddr_in *from;
{
    char *buf = NULL;
    int buf_len = sizeof(icpv1_4_common_t) + strlen(url) + 1;
    icpv1_4_common_t *headerp;
    icpv1_4UdpData *data = (icpv1_4UdpData *) xmalloc(sizeof(icpv1_4UdpData));
    struct sockaddr_in our_socket_name;
    int sock_name_length = sizeof(our_socket_name);

    if (getsockname(fd, (struct sockaddr *) &our_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "icpv1_4UdpMiss: FD %d: getsockname failure: %s\n",
	      fd, strerror(errno));
    }
    debug(5, "icpv1_4UdpMiss: FD %d: %s: <URL:%s>\n", fd,
	  inet_ntoa(our_socket_name.sin_addr), url);

    memset(data, '\0', sizeof(icpv1_4UdpData));
    memcpy(&data->address, from, sizeof(struct sockaddr_in));

    buf = (char *) xcalloc(1, buf_len);
    headerp = (icpv1_4_common_t *) buf;
    headerp->opcode = ICPV1_4_OP_MISS;
    headerp->version = ICPV1_4_VERSION_CURRENT;
    headerp->length = htons(buf_len);
    headerp->reqnum = htonl(reqheaderp->reqnum);
/*  memcpy(headerp->auth, , ICPV1_4_AUTH_SIZE); */
    headerp->shostid = htonl(our_socket_name.sin_addr.s_addr);

    memcpy(buf + sizeof(icpv1_4_common_t), url, strlen(url) + 1);
    data->msg = buf;
    data->len = buf_len;
    UdpQueue = AppendUdp(data, UdpQueue);

    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpv1_4UdpReply,
			    (caddr_t) UdpQueue);

    return COMM_OK;
}

int icpv1_4UdpSend(fd, url, reqheaderp, from, opcode)
int fd;
char *url;
icpv1_4_common_t *reqheaderp;
struct sockaddr_in *from;
icpv1_4_opcode opcode;
{
    char *buf;
    int buf_len = sizeof(icpv1_4_common_t) + strlen(url) + 1;
    icpv1_4_common_t *headerp;
    icpv1_4UdpData *data = (icpv1_4UdpData *) xmalloc(sizeof(icpv1_4UdpData));
    struct sockaddr_in our_socket_name;
    int sock_name_length = sizeof(our_socket_name);
    char *urloffset;

    if (getsockname(fd, (struct sockaddr *) &our_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "icpv1_4UdpSend: FD %d: getsockname failure: %s\n",
	      fd, strerror(errno));
    }
    memset(data, '\0', sizeof(icpv1_4UdpData));
    memcpy(&data->address, from, sizeof(struct sockaddr_in));

    if (opcode == ICPV1_4_OP_QUERY)
	buf_len += sizeof(u_num32);
    buf = (char *) xcalloc(1, buf_len);

    headerp = (icpv1_4_common_t *) buf;
    headerp->opcode = opcode;
    headerp->version = ICPV1_4_VERSION_CURRENT;
    headerp->length = htons(buf_len);
    headerp->reqnum = htonl(reqheaderp->reqnum);
/*  memcpy(headerp->auth, , ICPV1_4_AUTH_SIZE); */
    headerp->shostid = htonl(our_socket_name.sin_addr.s_addr);

    urloffset = buf + sizeof(icpv1_4_common_t);

    if (opcode == ICPV1_4_OP_QUERY)
	urloffset += sizeof(u_num32);
    /* it's already zero filled by xcalloc */
    memcpy(urloffset, url, strlen(url));
    data->msg = buf;
    data->len = buf_len;

    UdpQueue = AppendUdp(data, UdpQueue);

    debug(4, "icpv1_4UdpSend: op %d: from %s: sz %d: <URL:%s>\n", opcode,
	  inet_ntoa(from->sin_addr), buf_len, url);

    comm_set_select_handler(fd, COMM_SELECT_WRITE, (PF) icpv1_4UdpReply,
			    (caddr_t) UdpQueue);

    return COMM_OK;
}

int icpv1_4HandleUdp(sock, buf, from, len)
int sock;
caddr_t buf;
struct sockaddr_in *from;
int len;
{
    int result = 0;
    icpv1_4_common_t header, *headerp;
    StoreEntry *entry;
    char *url;

    if (len < sizeof(icpv1_4_common_t)) {
	debug(4, "icpv1_4HandleUdp: Bad UDP packet, Ignored. Size: %d < Header: %d\n",
	      len, sizeof(icpv1_4_common_t));
    } else {

	/* Get fields from incoming message. */
	headerp = (icpv1_4_common_t *) buf;
	header.opcode = headerp->opcode;
	header.length = ntohs(headerp->length);
	header.reqnum = ntohs(headerp->reqnum);
	header.shostid = ntohs(headerp->shostid);
	header.version = ntohs(headerp->version);;
	/*  memcpy(headerp->auth, , ICPV1_4_AUTH_SIZE); */

	switch (header.opcode) {
	case ICPV1_4_OP_QUERY:
	    {
		if (len < sizeof(icpv1_4_common_t)) {
		    /* at least it has to have \0 as a URL */
		    debug(4, "icpv1_4HandleUdp: Bad ICPV1_4_OP_QUERY packet, Ignored.\n");
		    debug(4, "Size: %d < Min: %d\n", len,
			  sizeof(icpv1_4_common_t) + sizeof(u_num32) + 1);
		    break;
		}
		url = buf + sizeof(header) + sizeof(u_num32);

		if (ip_access_check(from->sin_addr, ip_access_control_list) == IP_DENY) {
		    debug(2, "icpv1_4HandleUdp: Access Denied for %s.\n", inet_ntoa(from->sin_addr));
		    CacheInfo->log_append(CacheInfo, "ACCESS_DENIED", inet_ntoa(from->sin_addr), 0, "UDP_DENIED", "GET");
		} else {
		    entry = storeGet(url);
		    debug(5, "icpv1_4HandleUdp: OPCODE: ICPV1_4_OP_QUERY\n");

		    if (entry) {
			/* Send "HIT" message. */
			/* STAT */
			CacheInfo->log_append(CacheInfo, entry->url, inet_ntoa(from->sin_addr),
				    entry->object_len, "UDP_HIT", "GET");
			CacheInfo->proto_hit(CacheInfo, CacheInfo->proto_id(entry->url));

			icpv1_4UdpSend(sock, url, &header, from, ICPV1_4_OP_HIT);
		    } else {
			/* Send "MISS" message. */
			/* STAT */
			CacheInfo->log_append(CacheInfo, url, inet_ntoa(from->sin_addr),
					      0, "UDP_MISS", "GET");
			CacheInfo->proto_miss(CacheInfo, CacheInfo->proto_id(url));

			icpv1_4UdpMiss(sock, url, &header, from);
		    }
		}
	    }
	    break;

	case ICPV1_4_OP_HIT:
	    url = buf + sizeof(header);		/* XXX url known to be 0 term'd ? */
	    entry = storeGet(url);
	    debug(4, "icpv1_4HandleUdp: HIT for %s on <URL:%s>.\n", inet_ntoa(from->sin_addr), url);
	    if (!entry) {
		debug(4, "icpv1_4HandleUdp: received ICPV1_4_OP_HIT but StoreEntry is null. Ignored.\n");
	    } else {
		neighborsUdpAck(sock, url, &header, from, entry);
	    }
	    break;

	case ICPV1_4_OP_SECHO:
	    url = buf + sizeof(header);		/* XXX url known to be 0 term'd ? */
	    entry = storeGet(url);
	    debug(4, "icpv1_4HandleUdp: SECHO for %s on <URL:%s>\n", inet_ntoa(from->sin_addr), url);
	    if (!entry) {
		debug(4, "icpv1_4HandleUdp: received ICPV1_4_OP_SECHO but StoreEntry is null. Ignored.\n");
	    } else {
		neighborsUdpAck(sock, url, &header, from, entry);
	    }
	    break;

	case ICPV1_4_OP_MISS:
	    url = buf + sizeof(header);		/* XXX url known to be 0 term'd ? */
	    entry = storeGet(url);
	    debug(4, "icpv1_4HandleUdp: MISS for %s on <URL:%s>\n", inet_ntoa(from->sin_addr), url);
	    if (!entry) {
		debug(4, "icpv1_4HandleUdp: received ICPV1_4_OP_MISS but StoreEntry is null. Ignored.\n");
	    } else {
		neighborsUdpAck(sock, url, &header, from, entry);
	    }
	    break;

	default:
	    debug(4, "icpv1_4HandleUdp: OPCODE: %d\n", header.opcode);
	    break;
	}
    }
    return result;
}

#ifdef APPEND_DOMAIN
static char *
 do_append_domain(url)
char *url;
{
    char *b, *e, *p, *u;
    int lo, ln;

    if (!(b = strstr(url, "://")))
	return (url);
    b += 3;
    if (!(e = strchr(b, '/')))
	e = b + strlen(b);
    if ((p = strchr(b, '@')) && p < e)
	b = p + 1;
    if ((p = strchr(b, ':')) && p < e)
	e = p;
    if ((p = strchr(b, '.')) && p < e)
	return (url);
    lo = strlen(url);
    ln = lo + append_domain_len;
    if (!(u = xrealloc(url, ln + 1)))
	return (url);
    b = u + lo;
    p = u + ln;
    e = u + (e - url);
    while (b >= e)
	*p-- = *b--;
    strncpy(e, append_domain, append_domain_len);
    return (u);
}
#endif				/* APPEND_DOMAIN */

static
icpv1_4UdpData *
 AppendUdp(data, q)
icpv1_4UdpData *data;
icpv1_4UdpData *q;
{
    icpv1_4UdpData *p;

    data->next = NULL;
    if (q == NULL) {
	q = data;
	data->next = NULL;
    } else {
	p = q;
	while (p->next) {
	    p = p->next;
	}
	p->next = data;
    }
    return (q);
}

#endif /* SUPPORT_ICPV1_4 */
