static char rcsid[] = "$Id: objcache.c,v 1.35 1995/08/03 22:44:03 hardy Exp $";
/* 
 *  File:         objcache.c
 *  Description:  state machine for object_cache.
 *  Author:       Anawat Chankhunthod, USC
 *  Created:      May 12 1994
 *  Language:     C
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#include <sys/param.h>		/* For MAXHOSTNAMELEN. */
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "comm.h"
#include "store.h"
#include "stat.h"
#include "debug.h"
#include "cache_cf.h"

#define STAT_TTL 2

extern int errno;
extern char *sys_errlist[];

cacheinfo *CacheInfo;

typedef struct objcache_ds {
    StoreEntry *entry;
    char host[MAXHOSTNAMELEN];
    char request[1024];
    int reply_fd;
} ObjectCacheData;

/* user name for shutdown password in /etc/passwd */
char *username = "cache";
/* default password */
char *default_passwd = "V8AIBWT3k1MpQ";


/* Parse a object_cache url into components.  By Anawat. */
int objcache_url_parser(url, host, request, password)
char *host;
char *url;
char *request;
char *password;
{
    int t;

    host[0] = request[0] = password[0] = '\0';
    t = sscanf(url, "cache_object://%[^/]/%[^@]@%s", host, request, password);
    if (t < 2) {
	return -1;
    } else if (t == 2) {
	strcpy(password, "nopassword");
    }
    return 0;
}

int objcache_CheckPassword(password, user)
char *password;
char *user;
{
    struct passwd *pwd;
    char *salted_passwd;

    if (!password || !user)
	return -1;

    /* get password record from /etc/passwd */
    pwd = getpwnam(user);
    if (pwd)
	salted_passwd = pwd->pw_passwd;
    else
	salted_passwd = default_passwd;

    if (strcmp(salted_passwd, (char *) crypt(password, salted_passwd)) == 0)
	return 0;
    else
	return -1;


}

int objcacheStart(fd, url, entry)
int fd;
char *url;
StoreEntry *entry;
{
    int len = 0;
    char *buf = NULL;
    char *badentry = NULL;
    char *BADCacheURL = "Bad Object Cache URL %s ... negative cached.\n";
    char *BADPassword = "Incorrect password, sorry.\n";
    char password[64];
    struct sockaddr_in peer_socket_name;
    int sock_name_length = sizeof(peer_socket_name);

    /* Create state structure. */
    ObjectCacheData *data = (ObjectCacheData *) xmalloc(sizeof(ObjectCacheData));
    memset(data, '\0', sizeof(ObjectCacheData));
    data->reply_fd = fd;
    data->entry = entry;
    data->entry->ttl = STAT_TTL;	/* before we generate new object */

    debug(3, "objectcacheStart - url: %s\n", url);

    /* Parse url. */
    password[0] = '\0';
    if (objcache_url_parser(url, data->host, data->request, password)) {
	storeAbort(data->entry, "CACHED:OBJCACHE Invalid Syntax!\n");
	data->entry->ttl = STAT_TTL;	/* override negative TTL */
	safe_free(data);
	safe_free(buf);
	return COMM_ERROR;
    }
    if (getpeername(fd, (struct sockaddr *) &peer_socket_name,
		    &sock_name_length) == -1) {
	debug(1, "getpeername failed??\n");
    }
    if (ip_access_check(peer_socket_name.sin_addr, manager_ip_access_control_list)
	== IP_DENY) {		/* Access Deny */
	storeAbort(data->entry, "CACHED:OBJCACHE Access Denied!\n");
	data->entry->ttl = STAT_TTL;	/* override negative TTL */
	safe_free(data);
	safe_free(buf);
	return COMM_ERROR;
    }
    if (ipcache_checkname(comm_hostname(), data->host) != 0) {
	/* incorrect host name */
	storeAbort(data->entry, "CACHED:OBJCACHE Incorrect Hostname.\n");
	data->entry->ttl = STAT_TTL;	/* override negative TTL */
	safe_free(data);
	safe_free(buf);
	return COMM_ERROR;
    }
    /* retrieve object requested */
    if (strncmp(data->request, "shutdown", strlen("shutdown")) == 0) {
	if (objcache_CheckPassword(password, username) != 0) {
	    buf = strdup(BADPassword);
	    storeAppend(data->entry, buf, strlen(buf));
	    storeAbort(data->entry, "CACHED:OBJCACHE Incorrect Password\n");
	    /* override negative TTL */
	    data->entry->ttl = STAT_TTL;
	    debug(1, "Objcache: Attempt to shutdown cached with incorrect password\n");
	} else {
	    fprintf(debug_log_file, "Shutdown by command.\n");
	    /* free up state datastructure */
	    safe_free(data);
	    safe_free(buf);
	    PrintRusage(exit, debug_log_file);
	}

    } else if (strncmp(data->request, "info", strlen("info")) == 0) {
	CacheInfo->info_get(CacheInfo, &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "stats/objects", strlen("stats/objects")) == 0) {
	CacheInfo->stat_get(CacheInfo, "objects", &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "stats/utilization", strlen("stats/utilization")) == 0) {
	CacheInfo->stat_get(CacheInfo, "utilization", &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "stats/general", strlen("stats/general")) == 0) {
	CacheInfo->stat_get(CacheInfo, "general", &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "log/status", strlen("log/status")) == 0) {
	CacheInfo->log_status_get(CacheInfo, &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "log/enable", strlen("log/enable")) == 0) {
	CacheInfo->log_enable(CacheInfo, &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "log/disable", strlen("log/disable")) == 0) {
	CacheInfo->log_disable(CacheInfo, &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "log/clear", strlen("log/clear")) == 0) {
	CacheInfo->log_clear(CacheInfo, &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "log", strlen("log")) == 0) {
	CacheInfo->log_get_start(CacheInfo, data->entry);

    } else if (strncmp(data->request, "parameter", strlen("parameter")) == 0) {
	CacheInfo->parameter_get(CacheInfo, &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "server_list", strlen("server_list")) == 0) {
	CacheInfo->NotImplement(CacheInfo, &buf, &len);
	storeAppend(data->entry, buf, len);
	storeComplete(data->entry);

    } else if (strncmp(data->request, "cached.conf", strlen("cached.conf")) == 0) {
	CacheInfo->cached_get_start(CacheInfo, data->entry);

    } else {
	debug(5, "Bad Object Cache URL %s ... negative cached.\n", url);
	badentry = (char *) xcalloc(1, strlen(BADCacheURL) + strlen(url));
	sprintf(badentry, BADCacheURL, url);
	storeAppend(entry, badentry, strlen(badentry));
	safe_free(badentry);
	storeComplete(entry);
    }

    safe_free(data);
    safe_free(buf);
    return COMM_OK;
}
