/* 
 *  $Id: store.h,v 1.31 1995/08/21 07:54:55 duane Exp $
 *
 *  File:         store.h
 *  Description:  Interface to cache storage manager.
 *  Author:       John Noll, Anawat Chankhunthod, USC
 *  Created:      Sun Apr  3 16:51:36 1994
 *  Language:     C++
 **********************************************************************
 *  Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *          Mic Bowman of Transarc Corporation.
 *          Peter Danzig of the University of Southern California.
 *          Darren R. Hardy of the University of Colorado at Boulder.
 *          Udi Manber of the University of Arizona.
 *          Michael F. Schwartz of the University of Colorado at Boulder. 
 *          Duane Wessels of the University of Colorado at Boulder. 
 *  
 *  This copyright notice applies to all code in Harvest other than
 *  subsystems developed elsewhere, which contain other copyright notices
 *  in their source text.
 *  
 *  The Harvest software was developed by the Internet Research Task
 *  Force Research Group on Resource Discovery (IRTF-RD).  The Harvest
 *  software may be used for academic, research, government, and internal
 *  business purposes without charge.  If you wish to sell or distribute
 *  the Harvest software to commercial clients or partners, you must
 *  license the software.  See
 *  http://harvest.cs.colorado.edu/harvest/copyright,licensing.html#licensing.
 *  
 *  The Harvest software is provided ``as is'', without express or
 *  implied warranty, and with no support nor obligation to assist in its
 *  use, correction, modification or enhancement.  We assume no liability
 *  with respect to the infringement of copyrights, trade secrets, or any
 *  patents, and are not responsible for consequential damages.  Proper
 *  use of the Harvest software is entirely the responsibility of the user.
 *  
 *  For those who are using Harvest for non-commercial purposes, you may
 *  make derivative works, subject to the following constraints:
 *  
 *  - You must include the above copyright notice and these accompanying 
 *    paragraphs in all forms of derivative works, and any documentation 
 *    and other materials related to such distribution and use acknowledge 
 *    that the software was developed at the above institutions.
 *  
 *  - You must notify IRTF-RD regarding your distribution of the 
 *    derivative work.
 *  
 *  - You must clearly notify users that your are distributing a modified 
 *    version and not the original Harvest software.
 *  
 *  - Any derivative product is also subject to the restrictions of the 
 *    copyright, including distribution and use limitations.
 *  
 */
#ifndef _STORE_H_
#define _STORE_H_

#include <sys/param.h>

#include "proto.h"
#include "stmem.h"
#include "neighbors.h"
#include "ansihelp.h"

#define MAX_FILE_NAME_LEN 	256
#define MIN_PENDING 		1
#define IPPENDING_MAX 		10
#define MAX_POINTER_URLS	256 
#define BIT_SET(flag, bit) 	((flag) |= (bit))
#define BIT_RESET(flag, bit) 	((flag) &= ~(bit))
#define BIT_TEST(flag, bit) 	((flag) & (bit))

#define REQ_DISPATCHED 		(1<<17)
#define REQ_HTML 		(1<<16)
#define KEY_CHANGE 		(1<<15)
#define KEY_URL    		(1<<14)
#define CACHABLE   		(1<<11)
#define REFRESH_REQUEST   	(1<<10)
#define RELEASE_REQUEST 	(1<<9)
#define ABORT_MSG_PENDING 	(1<<8)
#define DELAY_SENDING 		(1<<7)
#define CLIENT_ABORT_REQUEST 	(1<<6)
#define OBJ_LOCK 		(1<<0)

/* type id for REQUEST opcode */
#define REQUEST_OP_GET     	0
#define REQUEST_OP_POST    	1
#define REQUEST_OP_HEAD    	2

extern char *HTTP_OPS[];

/* This sentry data structure is optimized to reduce space */
typedef struct sentry {
    char *url;
    char *key;
    char *mime_hdr;		/* Mime header info */
    u_num32 flag;
    enum {
	NOT_IN_MEMORY, SWAPPING_IN, IN_MEMORY
    } mem_status:2;
    enum {
	WAITING, TIMEOUT, DONE, NOPING
    } ping_status:2;
    enum {
	STORE_OK, STORE_PENDING, STORE_ABORTED
    } status:2;
    enum {
	NO_SWAP, SWAP_IN_PROGRESS, SWAP_OK
    } swap_status:2;
    enum {
	REQ_GET = 0, REQ_POST = 1, REQ_HEAD = 2
    } type_id:2;
    unsigned short swap_fd;
    mem_ptr data;

/* These items are mutually exclusive */
#define e_swap_buf crunch.swap_buf
#define e_pings_first_miss crunch.first_miss

    union {
	edge *first_miss;
	char *swap_buf;
    } crunch;

/* These items are also mutually exclusive */
#define e_swap_buf_len crunch2.swap_buf_len
#define e_pings_n_pings crunch2.pings.npings
#define e_pings_n_acks crunch2.pings.nacks
    union {
	struct {
	    unsigned char npings;
	    unsigned char nacks;
	} pings;
	int swap_buf_len;
    } crunch2;

#ifdef USE_DANZIG_OLD_CODE
#define e_swap_access crunch3.swap_access
#define e_abort_msg   crunch3.abort_msg
    union {
	int swap_access;
	char *abort_msg;
    } crunch3;
#else
#define e_swap_access no_union_swap_access
#define e_abort_msg   no_union_abort_msg
    int no_union_swap_access;
    char *no_union_abort_msg;
#endif

    int e_current_len;
    int object_len;
    u_num32 timestamp;
    u_num32 lastref;
    u_num32 refcount;
    u_num32 ttl;

    /* use another field to avoid changing the existing code */
    struct pentry **pending;
    unsigned char pending_list_size;

#ifdef VERY_RISKY_CODE
    u_num32 *pending_ip;
    unsigned char ipnum_pending;
#endif

    /* WARNING: Explicit assummption that fewer than 256
       WARNING:  clients all hop onto the same object.  The code
       WARNING:  doesn't deal with this case.
     */
    unsigned char lock_count;

    int swap_file_number;
    u_num32 swap_offset;

} StoreEntry;

typedef int (*PIF) _PARAMS((int, StoreEntry *, caddr_t));

typedef struct pentry {
    short fd;
    PIF handler;
    caddr_t data;
} PendingEntry;

/* 
 * XXX Storage for pointers, needs some refinement to allow this to have 
 * persistent data
 */ 
typedef struct ptr_
{
	char *ptr_url;	
	u_num32 ttl;
	struct ptr_ *next;

	/* XXX Some info for LRU management is needed (maybe)*/

} StorePointer;

typedef struct ptr_entry_
{
	char *url;
	unsigned short nr_ptrs;
	u_num32 timestamp; 
	int object_len;
	unsigned char local;
	char *hit_data;      /* HIT data to send */ 
	int hit_data_len;

	StorePointer *ptr;

	/* XXX Info for LRU management needed */
} StorePtrEntry;

extern StoreEntry *storeGet _PARAMS((char *));
extern StoreEntry *storeAdd _PARAMS((char *, char *, char *, int, int, int));
extern StoreEntry *storeGetFirst _PARAMS((void));
extern StoreEntry *storeGetNext _PARAMS((void));
extern StoreEntry *storeLRU _PARAMS((void));
extern int storeWalkThrough _PARAMS((int (*proc) (), caddr_t data));
extern void storePurgeOld _PARAMS((void));
extern void storeChangeKey _PARAMS((StoreEntry *));
extern void storeSanityCheck _PARAMS((char *));
extern int storeInit _PARAMS((char *));
extern int storeReleaseEntry _PARAMS((StoreEntry *));
extern int storeComplete _PARAMS((StoreEntry *));
extern int storeAbort _PARAMS((StoreEntry *, char *));
extern int storeAppend _PARAMS((StoreEntry *, char *, int));
extern int storeGetMemSize _PARAMS((void));
extern int storeGetMemSpace _PARAMS((int, int));
extern int storeGetSwapSize _PARAMS((void));
extern int storeGetSwapSpace _PARAMS((int));
extern int storeEntryValidToSend _PARAMS((StoreEntry *));
extern int storeEntryLocked _PARAMS((StoreEntry *));
extern int storeOriginalKey _PARAMS((StoreEntry *));
extern int storeGrep _PARAMS((StoreEntry *, char *, int));
extern char *storeGenerateKey _PARAMS((char *, int));
extern char *storeMatchMime _PARAMS((StoreEntry *, char *, char *, int));
extern StorePtrEntry *storeCheckCache _PARAMS((char *));

#endif
